import requests
import io
import gzip
import json

import pandas as pd
import scipy.io as sio
from matplotlib.image import imread
from anndata import AnnData

from decoupler._docs import docs
from decoupler._log import _log


@docs.dedent
def msvisium(
    verbose: bool = False,
) -> AnnData:
    """
    Downloads a spatial RNA-seq (Visium) human sample with multiple sclerosis
    displaying a chronic active lesion in the white matter of the brain :cite:`msvisium`.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    AnnData object.
    """
    url = (
        'https://www.ncbi.nlm.nih.gov/geo/download/'
        '?acc=GSM8563708&format=file&file=GSM8563708%5FMS377T%5F'
    )
    # Download mat
    response = requests.get(url + 'matrix%2Emtx%2Egz')
    with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as f:
        X = sio.mmread(f).T.tocsr().rint()
        X.eliminate_zeros()
    var = pd.read_csv(
        url + 'features%2Etsv%2Egz',
        compression='gzip',
        sep='\t',
        header=None,
        usecols=[1],
        index_col=0,
    )
    var.index.name = None
    # Remove repeated genes
    msk_var = ~(var.index.duplicated(keep='first'))
    var = var.loc[msk_var]
    X = X[:, msk_var]
    obs = pd.read_csv(
        url + 'barcodes%2Etsv%2Egz',
        compression='gzip',
        sep='\t',
        header=None,
        usecols=[0],
        index_col=0,
    )
    obs.index.name = None
    # Create anndata
    adata = AnnData(X=X, obs=obs, var=var)
    # Add images
    adata.uns['spatial'] = dict()
    adata.uns['spatial']['MS377T'] = dict()
    adata.uns['spatial']['MS377T']['images'] = dict()
    response = requests.get(url + 'scalefactors%5Fjson%2Ejson%2Egz')
    with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as f:
        adata.uns['spatial']['MS377T']['scalefactors'] = json.load(f)
    response = requests.get(url + 'tissue%5Fhires%5Fimage%2Epng%2Egz')
    with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as f:
        adata.uns['spatial']['MS377T']['images']['hires'] = imread(f)
    response = requests.get(url + 'tissue%5Flowres%5Fimage%2Epng%2Egz')
    with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as f:
        adata.uns['spatial']['MS377T']['images']['lowres'] = imread(f)
    # Add coordinates
    coords = pd.read_csv(
        url + 'tissue%5Fpositions%5Flist%2Ecsv%2Egz',
        compression='gzip',
        index_col=0,
    )
    adata.obs = adata.obs.join(coords, how='left')
    adata.obsm['spatial'] = adata.obs[['pxl_col_in_fullres', 'pxl_row_in_fullres']].values
    adata.obs.drop(
        columns=['in_tissue', 'pxl_row_in_fullres', 'pxl_col_in_fullres'],
        inplace=True,
    )
    # Add metadata
    url_meta = (
        'https://cells-test.gi.ucsc.edu/ms-subcortical-lesions/'
        'visium-ms377T/meta.tsv'
    )
    meta = pd.read_csv(url_meta, sep='\t', usecols=[0, 4], index_col=0)
    adata = adata[meta.index, :].copy()
    adata.obs = adata.obs.join(meta, how='right')
    adata.obs['niches'] = adata.obs['niches'].astype('category')
    adata.obs.index.name = None
    # Filter vars
    msk_var = adata.X.getnnz(axis=0) > 9
    adata = adata[:, msk_var].copy()
    m = f'generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata



import pandas as pd
import scipy.sparse as sps

import decoupler as dc


def test_rankby_order(
    tdata,
):
    df = dc.tl.rankby_order(tdata, order='pstime')
    assert isinstance(df, pd.DataFrame)
    neg_genes = {'G01', 'G02', 'G03', 'G04'}
    pos_genes = {'G05', 'G06', 'G07', 'G08'}
    gt_genes = neg_genes | pos_genes
    pd_genes = set(df[df['padj'] < 0.01]['name'])
    assert len(gt_genes) > 3
    assert gt_genes == pd_genes
    msk = df['name'].isin(gt_genes)
    assert df[~msk]['impr'].mean() < df[msk]['impr'].mean()
    assert (df[df['name'].isin(neg_genes)]['sign'] == -1).all()
    assert (df[df['name'].isin(pos_genes)]['sign'] == +1).all()
    tdata.X = sps.csr_matrix(tdata.X)
    df = dc.tl.rankby_order(tdata, order='pstime')
    assert isinstance(df, pd.DataFrame)
    pd_genes = set(df[df['padj'] < 0.01]['name'])
    assert len(gt_genes) > 3
    assert gt_genes == pd_genes
    msk = df['name'].isin(gt_genes)
    assert df[~msk]['impr'].mean() < df[msk]['impr'].mean()
    assert (df[df['name'].isin(neg_genes)]['sign'] == -1).all()
    assert (df[df['name'].isin(pos_genes)]['sign'] == +1).all()



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.fixture
def df():
    df = pd.DataFrame(
        data = [
            ['TF1', 1, 1, 5],
            ['TF2', 3, 1, 10],
            ['TF3', 4, 10, 10],
            ['TF4', 5, 15, 11],
        ],
        columns=['y', 'x', 'c', 's'],
    )
    return df


@pytest.mark.parametrize(
    'vcenter', [None, 3]
)
def test_dotplot(
    df,
    vcenter,
):
    fig = dc.pl.dotplot(df=df, x='x', y='y', c='c', s='s', vcenter=vcenter, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



import pandas as pd
import numpy as np
from anndata import AnnData
from matplotlib.figure import Figure

from decoupler._docs import docs
from decoupler._Plotter import Plotter


@docs.dedent
def filter_by_prop(
    adata: AnnData,
    min_prop: float = 0.1,
    min_smpls: int = 2,
    log: bool = True,
    color = 'gray',
    **kwargs
) -> None | Figure:
    """
    Plot to help determining the thresholds of the ``decoupler.pp.filter_by_prop`` function.

    Parameters
    ----------
    %(adata)s
    %(min_prop_prop)s
    %(min_smpls)s
    log
        Whether to log-scale the y axis.
    color
        Color to use in ``matplotlib.pyplot.hist``.
    %(plot)s
    """
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    assert 'psbulk_props' in adata.layers.keys(), \
    'psbulk_props must be in adata.layers, use this function afer running decoupler.pp.pseudobulk'
    props = adata.layers['psbulk_props']
    if isinstance(props, pd.DataFrame):
        props = props.values
    nsmpls = np.sum(props >= min_prop, axis=0)
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    _ = bp.ax.hist(
        nsmpls,
        bins=range(min(nsmpls), max(nsmpls) + 2),
        log=log,
        color=color,
        align='left',
        rwidth=0.95,
    )
    bp.ax.axvline(x=min_smpls - 0.5, c='black', ls='--')
    bp.ax.set_xlabel('Samples (≥ min_prop)')
    bp.ax.set_ylabel('Number of genes')
    return bp._return()



import pandas as pd
import pytest
import anndata as ad

import decoupler as dc


def test_hsctgfb():
    adata = dc.ds.hsctgfb()
    assert isinstance(adata, ad.AnnData)
    assert isinstance(adata.obs, pd.DataFrame)
    assert {'condition', 'sample_id'}.issubset(adata.obs)

@pytest.mark.parametrize(
    'thr_fc', [None, -1]
)
def test_knocktf(
    thr_fc, # val, None
):
    adata = dc.ds.knocktf(thr_fc=thr_fc)
    assert isinstance(adata, ad.AnnData)
    assert isinstance(adata.obs, pd.DataFrame)
    assert {'source', 'type_p'}.issubset(adata.obs.columns)
    if thr_fc is not None:
        assert (adata.obs['logFC'] < thr_fc).all()



import pytest

import decoupler as dc


def test_methods():
    lstm = dc.mt._methods
    len_lstm = len(lstm)
    len_dfm = dc.mt.show().shape[0]
    assert len_lstm == len_dfm
    assert all(isinstance(m, dc._Method.Method) for m in lstm)



import numpy as np
import scipy.stats as sts
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'a,b,c,d',
    [
        [10, 1, 2, 1000],
        [0, 20, 35, 5],
        [1, 2, 3, 4],
        [0, 1, 2, 500],
    ]
)
def test_table(
    a,
    b,
    c,
    d,
):
    dc_es = dc.mt._ora._oddsr(a=a, b=b, c=c, d=d, ha_corr=0., log=False)
    dc_pv = dc.mt._ora._test1t(a=a, b=b, c=c, d=d)
    st_es, st_pv = sts.fisher_exact([[a, b],[c, d]])
    assert np.isclose(dc_es, st_es)
    assert np.isclose(dc_pv, st_pv)


def test_func_ora(
    mat,
    idxmat,
):
    X, obs, var = mat
    cnct, starts, offsets = idxmat
    n_up = 3
    ha_corr = 1
    dc_es, dc_pv = dc.mt._ora._func_ora(
        mat=X,
        cnct=cnct,
        starts=starts,
        offsets=offsets,
        n_up=n_up,
        n_bm=0,
        n_bg=0,
        ha_corr=1,
    )
    st_es, st_pv = np.zeros(dc_es.shape), np.zeros(dc_pv.shape)
    ranks = np.arange(X.shape[1], dtype=np.int_)
    rnk = set(ranks)
    for i in range(st_es.shape[0]):
        row = sts.rankdata(X[i], method='ordinal')
        row = set(ranks[row > n_up])
        for j in range(st_es.shape[1]):
            fset = dc.pp.net._getset(cnct=cnct, starts=starts, offsets=offsets, j=j)
            fset = set(fset)
            # Build table
            set_a = row.intersection(fset)
            set_b = fset.difference(row)
            set_c = row.difference(fset)
            a = len(set_a)
            b = len(set_b)
            c = len(set_c)
            set_u = set_a.union(set_b).union(set_c)
            set_d = rnk.difference(set_u)
            d = len(set_d)
            _, st_pv[i, j] = sts.fisher_exact([[a, b],[c, d]])
            a += ha_corr
            b += ha_corr
            c += ha_corr
            d += ha_corr
            es = sts.fisher_exact([[a, b],[c, d]])
            st_es[i, j], _ = np.log(es)
    assert np.isclose(dc_es, st_es).all()
    assert np.isclose(dc_pv, st_pv).all()



import logging

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import to_hex
from matplotlib.figure import Figure
import marsilea as ma
import marsilea.plotter as mp

from decoupler._docs import docs
from decoupler._Plotter import Plotter


def _mcolor(
    cmap: str,
) -> str:
    assert isinstance(cmap, str), 'cmap must be str'
    cmap = plt.get_cmap(cmap)
    mid_color_rgb = cmap(0.5)
    mid_color_hex = to_hex(mid_color_rgb)
    return mid_color_hex


def _palette(
    labels: list,
    cmap: str = 'tab20'
) -> dict:
    assert isinstance(cmap, str), 'cmap must be str'
    rnks = np.arange(len(labels))
    if cmap in plt.colormaps():
        cmap = plt.get_cmap(cmap)
        sidx = np.argsort(np.argsort(labels))
        colors = np.array([to_hex(cmap(i)) for i in rnks])
        palette = dict(zip(rnks, colors[sidx]))
    else:
        palette = dict(zip(rnks, [cmap for _ in range(rnks.size)]))
    return palette


@docs.dedent
def summary(
    df: pd.DataFrame,
    y: str,
    metrics: list = ['auc', 'fscore', 'qrank'],
    cmap_y: str = 'tab20',
    cmap_auc: str = 'Greens',
    cmap_fscore: str = 'Blues',
    cmap_qrank: str = 'Reds',
    **kwargs
) -> None | Figure:
    """
    Summarizes metrics into a final score by computing the quantile-normalized rank across them.

    Parameters
    ----------
    %(df)s
    %(y)s
    metrics
        Which metrics to include.
    cmap_y
        Color map for the y grouping.
    cmap_auc
        Color map for the auc metric.
    cmap_fscore
        Color map for the fscore metric.
    cmap_qrank
        Color map for the qrank metric.
    %(plot)s
    """
    # Validate
    assert isinstance(df, pd.DataFrame), 'df must be pandas.DataFrame'
    assert isinstance(y, str) and y in df.columns, 'y must be str and in df.columns'
    assert isinstance(metrics, (str, list)), 'metrics must be str or list'
    if isinstance(metrics, str):
        metrics = [metrics]
    assert set(metrics).issubset({'auc', 'fscore', 'qrank'})
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    # Fix
    df = df.copy().sort_values('score', ascending=False).set_index(y)
    # Scale
    cols = list(df.select_dtypes(include='number').columns)
    df[cols] = (df[cols] - df[cols].min()) / (df[cols].max() - df[cols].min())
    # Instance
    kwargs['ax'] = None
    bp = Plotter(**kwargs)
    bp.fig.delaxes(bp.ax)
    plt.close(bp.fig)
    # Plot
    lst_h = []
    pad = 0.05
    if 'auc' in metrics:
        cnames = ['auroc', 'auprc', 'H(auroc, auprc)']
        h = ma.Heatmap(df[cnames], cmap=cmap_auc, name="h_auc", width=1, height=3, lw=.1, cbar_kws=dict(orientation="horizontal", title="Scaled auc"))
        h.group_cols(['auc', 'auc', 'auc'], order=['auc'])
        h.add_top(
            ma.plotter.Chunk(['auc'], fill_colors=[_mcolor(cmap_auc)], align='center')
        )
        h.add_bottom(mp.Labels(cnames[:2] + ['h mean'], ), pad=pad)
        lst_h.append(h)
    if 'fscore' in metrics:
        cnames = ['precision', 'recall', 'F-score']
        h = ma.Heatmap(df[cnames], cmap=cmap_fscore, name="h_fsc", width=1, height=2, lw=.1, cbar_kws=dict(orientation="horizontal", title="Scaled fscore"))
        h.group_cols(['fscore', 'fscore', 'fscore'], order=['fscore'])
        h.add_top(
            ma.plotter.Chunk(['fscore'], fill_colors=[_mcolor(cmap_fscore)], align='center')
        )
        h.add_bottom(mp.Labels(cnames[:2] + ['h mean'], ), pad=pad)
        lst_h.append(h)
    if 'qrank' in metrics:
        cnames = ['-log10(pval)', '1-qrank', 'H(1-qrank, -log10(pval))']
        h = ma.Heatmap(df[cnames], cmap=cmap_qrank, name="h_rnk", width=1, height=1, lw=.1, cbar_kws=dict(orientation="horizontal", title="Scaled qrank"))
        h.group_cols(['qrank', 'qrank', 'qrank'], order=['qrank'])
        h.add_top(
            ma.plotter.Chunk(['qrank'], fill_colors=[_mcolor(cmap_qrank)], align='center')
        )
        h.add_bottom(mp.Labels(cnames[:2] + ['h mean'], ), pad=pad)
        lst_h.append(h)
    # Build canvas
    c = lst_h[0]
    # Add bar to the left
    bar = mp.Bar(df['score'].T, label='Score', color='gray', palette=_palette(labels=df.index, cmap=cmap_y))
    c.add_left(bar)
    c.add_left(mp.Labels(df.index, ), pad=pad)
    c.add_title(left=y, pad=pad)
    for h in lst_h[1:]:
        c += h
    c.add_legends(side='right')
    c.render()
    if bp.return_fig or bp.save is not None:
        plt.close()
    i = [i for i in range(len(c.figure.axes)) if c.figure.axes[i].get_xlabel() == 'Score'][0]
    c.figure.axes[i].invert_xaxis()
    bp.fig = c.figure
    bp.fig.set_figwidth(bp.figsize[0])
    bp.fig.set_figheight(bp.figsize[1])
    bp.fig.set_dpi(bp.dpi)
    return bp._return()



import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s | [%(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S"
)

def _log(
    message: str,
    level: str = 'info',
    verbose: bool = False
) -> None:
    """
    Log a message with a specified logging level.

    Parameters
    ----------
    message
        The message to log.
    level
        The logging level.
    verbose
        Whether to emit the log.
    """
    level = level.lower()
    if verbose:
        if level == "warn":
            logging.warning(message)
        elif level == "info":
            logging.info(message)



import pandas as pd
import numpy as np
from matplotlib.figure import Figure
import adjustText as at

from decoupler._Plotter import Plotter
from decoupler.pp.net import _validate_net


def volcano(
    data: pd.DataFrame,
    x: str,
    y: str,
    net: pd.DataFrame | None = None,
    name: str | None = None,
    top: int = 5,
    thr_stat: float = 0.5,
    thr_sign: float = 0.05,
    max_stat: float | None = None,
    max_sign: float | None = None,
    color_pos: str = '#D62728',
    color_neg: str = '#1F77B4',
    color_null: str = 'gray',
    **kwargs
) -> None | Figure:
    """
    Plot logFC and p-values from a long formated data-frame.

    Parameters
    ----------
    %(data_plot)s
    x
        Column name of ``data`` storing the change statitsics.
    y
        Column name of ``data`` storing the associated p-values.
    %(net)s
    name
        Name of the source to subset ``net``.
    top
        Number of top differentially abundant features to show.
    thr_stat
        Significance threshold for change statitsics.
    thr_sign
        Significance threshold for p-values.
    max_stat
        Limit of change statitsics to plot in absolute value.
    max_sign
        Limit of p-values to plot in ``-log10``.
    color_pos
        Color to plot significant positive features.
    color_neg
        Color to plot significant negative features.
    color_null
        Color to plot rest of the genes.
    %(plot)s
    """
    # Validate inputs
    m = f'data must be a pd.DataFrame containing the columns {x} and {y}'
    assert isinstance(data, pd.DataFrame), m
    assert {x, y}.issubset(data.columns), m
    assert (net is None) == (name is None), \
    'net and name must be both defined or both None'
    assert isinstance(top, int) and top > 0, 'top must be int and > 0'
    assert isinstance(thr_stat, (int, float)) and thr_stat > 0, \
    'thr_stat must be numeric and > 0'
    assert isinstance(thr_sign, (int, float)) and thr_sign > 0, \
    'thr_sign must be numeric and > 0'
    if max_stat is None:
        max_stat = np.inf
    if max_sign is None:
        max_sign = np.inf
    assert isinstance(max_stat, (int, float)) and max_stat > 0, \
    'max_stat must be None, or numeric and > 0'
    assert isinstance(max_sign, (int, float)) and max_sign > 0, \
    'max_sign must be None, or numeric and > 0'
    assert isinstance(color_pos, str), 'color_pos must be str'
    assert isinstance(color_neg, str), 'color_neg must be str'
    assert isinstance(color_null, str), 'color_null must be str'
    # Instance
    bp = Plotter(**kwargs)
    # Transform thr_sign
    thr_sign = -np.log10(thr_sign)
    # Extract df
    df = data.copy()
    df['stat'] = df[x]
    non_zero_min = df[y][df[y] != 0].min()
    df['pval'] = -np.log10(df[y].clip(lower=non_zero_min, upper=1))
    # Filter by net shared targets
    if net is not None:
        vnet = _validate_net(net)
        snet = vnet[vnet['source'] == name]
        df = df[df.index.isin(snet['target'])]
    # Filter by limits
    msk_stat = np.abs(df['stat']) < np.abs(max_stat)
    msk_sign = df['pval'] < np.abs(max_sign)
    df = df.loc[msk_stat & msk_sign]
    # Define color by up or down regulation and significance
    df['weight'] = color_null
    up_msk = (df['stat'] >= thr_stat) & (df['pval'] >= thr_sign)
    dw_msk = (df['stat'] <= -thr_stat) & (df['pval'] >= thr_sign)
    df.loc[up_msk, 'weight'] = color_pos
    df.loc[dw_msk, 'weight'] = color_neg
    # Plot
    df.plot.scatter(x='stat', y='pval', c='weight', sharex=False, ax=bp.ax)
    bp.ax.set_axisbelow(True)
    # Draw thr lines
    bp.ax.axvline(x=thr_stat, linestyle='--', color="black")
    bp.ax.axvline(x=-thr_stat, linestyle='--', color="black")
    bp.ax.axhline(y=thr_sign, linestyle='--', color="black")
    # Add labels
    bp.ax.set_title(name)
    bp.ax.set_xlabel(x)
    bp.ax.set_ylabel(rf'$-\log_{{10}}({y})$')
    # Show top sign features
    signs = df[up_msk | dw_msk].sort_values('pval', ascending=False)
    signs = signs.iloc[:top]
    texts = []
    for x, y, s in zip(signs['stat'], signs['pval'], signs.index):
        texts.append(bp.ax.text(x, y, s))
    if len(texts) > 0:
        at.adjust_text(texts, arrowprops=dict(arrowstyle='-', color='black'), ax=bp.ax)
    return bp._return()



from typing import Tuple

import numpy as np
import scipy.stats as sts

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method


def _cov(
    A: np.ndarray,
    b: np.ndarray
) -> np.ndarray:
    return np.dot(b.T - b.mean(), A - A.mean(axis=0)) / (b.shape[0]-1)


def _cor(
    A: np.ndarray,
    b: np.ndarray
) -> np.ndarray:
    cov = _cov(A, b)
    ssd = np.std(A, axis=0, ddof=1) * np.std(b, axis=0, ddof=1).reshape(-1, 1)
    return cov / ssd


def _tval(
    r: np.ndarray,
    df: float
) -> np.ndarray:
    return r * np.sqrt(df / ((1.0 - r + 2.2e-16) * (1.0 + r + 2.2e-16)))


@docs.dedent
def _func_ulm(
    mat: np.ndarray,
    adj: np.ndarray,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Univariate Linear Model (ULM) :cite:`decoupler`.

    This approach uses the molecular features from one observation as the population of samples
    and it fits a linear model with a single covariate, which is the feature weights of a set :math:`F`.

    .. math::

        y_i = \beta_0 + \beta_1 x_i + \varepsilon, \quad i = 1, 2, \ldots, n

    Where:

    - :math:`y_i` is the observed feature statistic (e.g. gene expression, :math:`log_{2}FC`, etc.) for feature :math:`i`
    - :math:`x_i` is the weight of feature :math:`i` in feature set :math:`F`. For unweighted sets, membership in the set is indicated by 1, and non-membership by 0.
    - :math:`\beta_0` is the intercept
    - :math:`\beta_1` is the slope coefficient
    - :math:`\varepsilon` is the error term for feature :math:`i`

    .. figure:: /_static/images/ulm.png
       :alt: Univariate Linear Model (ULM) schematic.
       :align: center
       :width: 75%

       Univariate Linear Model (ULM) scheme.
       In this example, the observed gene expression of :math:`Sample_1` is predicted using
       the interaction weights of :math:`TF_1`.
       Since the target genes that have negative weights are lowly expressed,
       and the positive target genes are highly expressed,
       the relationship between the two variables is positive so the obtained :math:`ES` score is positive.
       Scores can be interpreted as active when positive, repressive when negative, and inconclusive when close to 0.

    The enrichment score :math:`ES` is then calculated as the t-value of the slope coefficient.

    .. math::

        ES = t_{\beta_1} = \frac{\hat{\beta}_1}{\mathrm{SE}(\hat{\beta}_1)}

    Where:

    - :math:`t_{\beta_1}` is the t-value of the slope
    - :math:`\mathrm{SE}(\hat{\beta}_1)` is the standard error of the slope

    Next, :math:`p_{value}` are obtained by evaluating the two-sided survival function
    (:math:`sf`) of the Student’s t-distribution.

    .. math::

        p_{value} = 2 \times \mathrm{sf}(|ES|, \text{df})

    %(yestest)s

    %(params)s

    %(returns)s
    """
    # Get degrees of freedom
    n_var, n_src = adj.shape
    df = n_var - 2
    m = f'ulm - fitting {n_src} univariate models of {n_var} observations (targets) with {df} degrees of freedom'
    _log(m, level='info', verbose=verbose)
    # Compute R value for all
    r = _cor(adj, mat.T)
    # Compute t-value
    es = _tval(r, df)
    # Compute p-value
    pv = sts.t.sf(abs(es), df) * 2
    return es, pv


_ulm = MethodMeta(
    name='ulm',
    desc='Univariate Linear Model (ULM)',
    func=_func_ulm,
    stype='numerical',
    adj=True,
    weight=True,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1093/bioadv/vbac016',
)
ulm = Method(_method=_ulm)



from typing import Tuple
import math

import numpy as np
import scipy.stats as sts
import scipy.sparse as sps
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method
from decoupler.pp.net import _getset


def _maxn() -> int:
    l = 1; n = 2; h = float('inf')
    while l < n:
        if abs(math.lgamma(n+1) - math.lgamma(n) - math.log(n)) >= 1: h = n
        else: l = n
        n = (l + min(h, l * 3)) // 2
    return n

MAXN = _maxn()

@nb.njit(cache=True)
def _mlnTest2t(
    a: int,
    ab: int,
    ac: int,
    abcd: int,
):
    if 0 > a or a > ab or a > ac or ab + ac > abcd + a: raise ValueError('invalid contingency table')
    if abcd > MAXN: raise OverflowError('the grand total of contingency table is too large')
    a_min = max(0, ab + ac - abcd)
    a_max = min(ab, ac)
    if a_min == a_max: return 0.
    p0 = math.lgamma(ab + 1) + math.lgamma(ac + 1) + math.lgamma(abcd - ac + 1) + math.lgamma(abcd - ab + 1) - math.lgamma(abcd + 1)
    pa = math.lgamma(a + 1) + math.lgamma(ab - a + 1) + math.lgamma(ac - a + 1) + math.lgamma(abcd - ab - ac + a + 1)
    st = 1.
    if ab * ac < a * abcd:
        for i in range(min(a - 1, int(round(ab * ac / abcd))), a_min - 1, -1):
            pi = math.lgamma(i + 1) + math.lgamma(ab - i + 1) + math.lgamma(ac - i + 1) + math.lgamma(abcd - ab - ac + i + 1)
            if pi < pa: continue
            st_new = st + math.exp(pa - pi)
            if st_new == st: break
            st = st_new
        for i in range(a + 1, a_max + 1):
            pi = math.lgamma(i + 1) + math.lgamma(ab - i + 1) + math.lgamma(ac - i + 1) + math.lgamma(abcd - ab - ac + i + 1)
            st_new = st + math.exp(pa - pi)
            if st_new == st: break
            st = st_new
    else:
        for i in range(a - 1, a_min - 1, -1):
            pi = math.lgamma(i + 1) + math.lgamma(ab - i + 1) + math.lgamma(ac - i + 1) + math.lgamma(abcd - ab - ac + i + 1)
            st_new = st + math.exp(pa - pi)
            if st_new == st: break
            st = st_new
        for i in range(max(a + 1, int(round(ab * ac / abcd))), a_max + 1):
            pi = math.lgamma(i + 1) + math.lgamma(ab - i + 1) + math.lgamma(ac - i + 1) + math.lgamma(abcd - ab - ac + i + 1)
            if pi < pa: continue
            st_new = st + math.exp(pa - pi)
            if st_new == st: break
            st = st_new
    return max(0, pa - p0 - math.log(st))


@nb.njit(cache=True)
def _test1t(
    a: int,
    b: int,
    c: int,
    d: int,
) -> float:
    # https://github.com/painyeph/FishersExactTest/blob/master/fisher.py
    return math.exp(-_mlnTest2t(a, a + b, a + c, a + b + c + d))


@nb.njit(cache=True)
def _oddsr(
    a: int,
    b: int,
    c: int,
    d: int,
    ha_corr: int | float = 0.5,
    log: bool = True,
):
    # Haldane-Anscombe correction
    a += ha_corr
    b += ha_corr
    c += ha_corr
    d += ha_corr
    r = (a * d) / (b * c)
    if log and r != 0.:
        r = math.log(r)
    return r


@nb.njit(parallel=True, cache=True)
def _runora(
    row: np.ndarray,
    ranks: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    n_bg: int | None,
    ha_corr: int | float = 0.5,
) -> Tuple[float, float]:
    nvar = row.size
    nsrc = starts.size
    # Transform to set
    row = set(row)
    ranks = set(ranks)
    es = np.zeros(nsrc)
    pv = np.zeros(nsrc)
    for j in nb.prange(nsrc):
        # Extract feature set
        fset = _getset(cnct=cnct, starts=starts, offsets=offsets, j=j)
        fset = set(fset)
        # Build table
        set_a = row.intersection(fset)
        set_b = fset.difference(row)
        set_c = row.difference(fset)
        a = len(set_a)
        b = len(set_b)
        c = len(set_c)
        if n_bg == 0:
            set_u = set_a.union(set_b).union(set_c)
            set_d = ranks.difference(set_u)
            d = len(set_d)
        else:
            d = n_bg - a - b - c
        es[j] = _oddsr(a=a, b=b, c=c, d=d, ha_corr=ha_corr, log=True)
        pv[j] = _test1t(a=a, b=b, c=c, d=d)
    return es, pv


@docs.dedent
def _func_ora(
    mat: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    n_up: int | float | None = None,
    n_bm: int | float = 0,
    n_bg: int | float | None = 20_000,
    ha_corr: int | float = 0.5,
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Over Representation Analysis (ORA) :cite:`ora`.

    This approach first creates a contingency table.

    .. list-table:: 2×2 Contingency Table
       :header-rows: 1
       :widths: 20 20 20

       * -
         - :math:`\in F`
         - :math:`\notin F`
       * - :math:`\in Sign. features`
         - :math:`a`
         - :math:`b`
       * - :math:`\notin Sign. features`
         - :math:`c`
         - :math:`d`

    Where:

    - :math:`a` is the number of features that are both significant and in :math:`F`
    - :math:`b` is the number of features that are signficiant but not in :math:`F`
    - :math:`c` is the number of features that are not signficiant but in :math:`F`
    - :math:`d` is the number of features that are not signficiant and not in :math:`F`

    .. figure:: /_static/images/ora.png
       :alt: Over Representation Analysis (ORA) schematic.
       :align: center
       :width: 75%

       Over Representation Analysis (ORA) scheme.
    
    The statistic is calculated as the Odds Ratio :math:`OR` with Haldane-Anscombe correction.

    .. math::
        
        \text{OR} = \log{\frac{\frac{a + 0.5}{b + 0.5}}{\frac{c + 0.5}{d + 0.5}}}

    And the :math:`p_{value}` is obtained afer computing a two-tailed Fisher’s exact test with the same table.

    %(yestest)s

    %(params)s

    n_up
        Number of top-ranked features, based on their magnitude, to select as observed features.
        If ``None``, the top 5% of positive features are selected.
    n_bm
        Number of bottom-ranked features, based on their magnitude, to select as observed features.
    n_bg
        Number indicating the background size.

    %(returns)s
    """
    nobs, nvar = mat.shape
    nsrc = starts.size
    if n_up is None:
        n_up = int(np.max([np.ceil(0.05 * nvar), 2]))
        m = f'ora - setting n_up={n_up}' 
        _log(m, level='info', verbose=verbose)
    if n_bg is None:
        n_bg = 0
        m = f'ora - not using n_bg, a feature specific background will be used instead' 
        _log(m, level='info', verbose=verbose)
    assert isinstance(n_up, (int, float)) and n_up > 0, 'n_up must be numeric and > 0'
    assert isinstance(n_bm, (int, float)) and n_bm >= 0, 'n_bm must be numeric and positive'
    assert isinstance(n_bg, (int, float)) and n_bg >= 0, 'n_bg must be numeric and positive'
    m = f'ora - calculating {nsrc} scores across {nobs} observations with n_up={n_up}, n_bm={n_bm}, n_bg={n_bg}' 
    _log(m, level='info', verbose=verbose)
    es = np.zeros((nobs, nsrc))
    pv = np.zeros((nobs, nsrc))
    ranks = np.arange(nvar, dtype=np.int_)
    for i in tqdm(range(nobs), disable=not verbose):
        if isinstance(mat, sps.csr_matrix):
            row = mat[i].toarray()[0]
        else:
            row = mat[i]
        # Find ranks
        row = sts.rankdata(row, method='ordinal')
        row = ranks[(row > n_up) | (row < n_bm)]
        es[i], pv[i] = _runora(row=row, ranks=ranks, cnct=cnct, starts=starts, offsets=offsets, n_bg=n_bg, ha_corr=ha_corr)
    return es, pv


_ora = MethodMeta(
    name='ora',
    desc='Over Representation Analysis (ORA)',
    func=_func_ora,
    stype='categorical',
    adj=False,
    weight=False,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.2307/2340521',
)
ora = Method(_method=_ora)



from decoupler.bm.metric._auc import auc
from decoupler.bm.metric._fscore import fscore
from decoupler.bm.metric._qrank import qrank
from decoupler.bm.metric._hmean import hmean

dict_metric = {
    'auc': auc,
    'fscore': fscore,
    'qrank': qrank,
}



from typing import Tuple

import numpy as np
import pandas as pd
import scipy.stats as sts
import numba as nb
from anndata import AnnData

from decoupler._docs import docs
from decoupler._log import _log
from decoupler.mt._gsea import _std
from decoupler.mt._run import _return


@nb.njit(cache=True)
def _zscore(
    sel: np.ndarray,
) -> np.ndarray:
    # Skip if all zeros
    if np.all(sel == 0):
        return sel
    # N selection
    n_sel = len(sel)
    # Mirror distr
    sel = np.append(sel, -sel)
    # Compute z-score (mean is 0)
    z = (sel / _std(sel, 1))[:n_sel]
    return z


@nb.njit(parallel=True, cache=True)
def _mean_zscores(
    scores: np.ndarray,
) -> np.ndarray:
    # Make a copy not to overwrite
    x = scores.copy()
    # Extract dims
    n_methods, n_samples, n_ftrs = x.shape
    # Init empty cons scores
    cons = np.zeros((n_samples, n_ftrs), dtype=nb.f4)
    # For each sample
    for i in nb.prange(n_samples):
        sample = x[:, i, :]
        # Compute z-score per method
        for k in range(n_methods):
            # Extract method scores
            methd = sample[k, :]
            # Compute pos z-score
            msk = methd > 0
            if np.any(msk):
                methd[msk] = _zscore(methd[msk])
            # Compute neg z-score
            msk = methd <= 0
            if np.any(msk):
                methd[msk] = _zscore(methd[msk])
        # Compute mean per feature
        for j in range(n_ftrs):
            ftr = sample[:, j]
            cons[i, j] = np.mean(ftr)
    return cons


def consensus(
    result: dict | AnnData,
    verbose: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame] | None:
    """
    Consensus score across methods.

    Parameters
    ----------
    result
        Results from ``decoupler.mt.decouple``.
    %(verbose)s

    Returns
    -------
    Consensus enrichment scores and p-values.
    """
    # Validate
    assert isinstance(result, (dict, AnnData)), 'scores must be dict or anndata.AnnData'
    # Transform to mat
    if isinstance(result, AnnData):
        keys = [k for k in result.obsm if 'score_' in k]
        scores = [result.obsm[k].values for k in keys]
        obs_names = result.obs_names
        var_names = result.obsm[keys[0]].columns
    else:
        keys = [k for k in result if 'score_' in k]
        scores = [result[k].values for k in keys]
        obs_names = result[keys[0]].index
        var_names = result[keys[0]].columns
    names = {k.split('_')[1] for k in keys}
    m = f'consensus - running consensus using methods={names}'
    _log(m, level='info', verbose=verbose)
    scores = np.array(scores)
    # Compute mean z-scores
    es = _mean_zscores(scores)
    # Compute p-vals
    pv = 2 * sts.norm.sf(np.abs(es))
    # FDR
    pv = sts.false_discovery_control(pv, axis=1, method='bh')
    # Transform to df
    es = pd.DataFrame(es, columns=var_names, index=obs_names)
    pv = pd.DataFrame(pv, columns=var_names, index=obs_names)
    return _return(name='consensus', data=result, es=es, pv=pv, verbose=False)



import numpy as np
import pandas as pd
import pytest
import scanpy as sc

import decoupler as dc


@pytest.fixture
def rng():
    rng = np.random.default_rng(seed=42)
    return rng


@pytest.fixture
def adata():
    adata, _ = dc.ds.toy(nobs=40, nvar=20, bval=2, seed=42, verbose=False)
    adata.layers['counts'] = adata.X.round()
    return adata


@pytest.fixture
def tdata():
    tdata, _ = dc.ds.toy(nobs=40, nvar=20, bval=2, seed=42, verbose=False, pstime=True)
    return tdata


@pytest.fixture
def tdata_obsm(
    tdata,
    net,
    rng,
):
    sc.tl.pca(tdata)
    tdata.obsm['X_umap'] = tdata.obsm['X_pca'][:, :2] + rng.random(tdata.obsm['X_pca'][:, :2].shape)
    dc.mt.ulm(data=tdata, net=net, tmin=0)
    return tdata


@pytest.fixture
def pdata(
    adata,
    rng,
):
    adata.X = adata.X.round() * (rng.random(adata.shape) > 0.75)
    return dc.pp.pseudobulk(adata=adata, sample_col='sample', groups_col='group')


@pytest.fixture
def deg():
    deg = pd.DataFrame(
        data = [
            [1, 0.5],
            [-2, 0.25],
            [3, 0.125],
            [-4, 0.05],
            [5, 0.025],
        ],
        columns=['stat', 'padj'],
        index=['G01', 'G02', 'G03', 'G04', 'G05']
    )
    return deg


@pytest.fixture
def net():
    _, net = dc.ds.toy(nobs=2, nvar=12, bval=2, seed=42, verbose=False)
    net = dc.pp.prune(features=net['target'].unique(), net=net, tmin=3)
    return net


@pytest.fixture
def unwnet(net):
    return net.drop(columns=['weight'], inplace=False)


@pytest.fixture
def mat(
    adata,
):
    return dc.pp.extract(data=adata)


@pytest.fixture
def idxmat(
    mat,
    net,
):
    X, obs, var = mat
    sources, cnct, starts, offsets = dc.pp.idxmat(features=var, net=net, verbose=False)
    return cnct, starts, offsets


@pytest.fixture
def adjmat(
    mat,
    net,
):
    X, obs, var = mat
    sources, targets, adjmat = dc.pp.adjmat(features=var, net=net, verbose=False)
    return adjmat



import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'lst_ens,lst_sym',
    [
        [['ENSG00000196092', 'ENSG00000115415'], ['PAX5', 'STAT1']],
        [['ENSG00000204655', 'ENSG00000184221'], ['MOG', 'OLIG1']],
    ]
)
def test_ensmbl_to_symbol(
    lst_ens,
    lst_sym,
):
    lst_trn = dc.ds.ensmbl_to_symbol(genes=lst_ens)
    assert all(s == t for s, t in zip(lst_trn, lst_sym))



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure

import decoupler as dc


def test_source_targets(
    deg,
    net,
):
    fig = dc.pl.source_targets(data=deg, net=net, name='T1', x='weight', y='stat', return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



from decoupler._Method import _show_methods
from decoupler.mt._methods import aucell
from decoupler.mt._methods import gsea
from decoupler.mt._methods import gsva
from decoupler.mt._methods import mdt
from decoupler.mt._methods import mlm
from decoupler.mt._methods import ora
from decoupler.mt._methods import udt
from decoupler.mt._methods import ulm
from decoupler.mt._methods import viper
from decoupler.mt._methods import waggr
from decoupler.mt._methods import zscore
from decoupler.mt._methods import _methods
from decoupler.mt._decouple import decouple
from decoupler.mt._consensus import consensus

def show() -> None:
    """Displays the methods available in decoupler"""
    return _show_methods(_methods)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.fixture
def data():
    data = pd.DataFrame(
        data=[
            [5, 6, 7, 1, 1, 2.],
        ],
        index=['C1'],
        columns=['G01', 'G02', 'G03', 'G06', 'G07', 'G08']
    )
    return data


@pytest.fixture
def score():
    score = pd.DataFrame(
        data=[
            [4, 3, -3, -2.],
        ],
        index=['C1'],
        columns=[f'T{i + 1}' for i in range(4)]
    )
    return score


@pytest.mark.parametrize(
    'd_none,unw,sources,targets,by_abs,vcenter',
    [
        [False, False, 5, 5, False, False],
        [False, True, 'T1', 5, True, True],
        [True, False, ['T1'], 5, True, True],
        [True, False, ['T1', 'T3'], 5, True, True],
        [False, False, 5, 'G01', True, True],
        [False, False, 5, ['G01', 'G02', 'G03'], True, True],
    ]
)
def test_network(
    net,
    data,
    score,
    d_none,
    unw,
    sources,
    targets,
    by_abs,
    vcenter,
):
    if d_none:
        s_cmap = 'white'
        data = None
        score = None
    else:
        s_cmap = 'coolwarm'
    if unw:
        net = net.drop(columns=['weight'])
    fig = dc.pl.network(
        data=data,
        score=score,
        net=net,
        sources=sources,
        targets=targets,
        by_abs=by_abs,
        vcenter=vcenter,
        s_cmap = s_cmap,
        figsize=(5, 5),
        return_fig=True
    )
    assert isinstance(fig, Figure)
    plt.close(fig)


from decoupler._Method import _show_methods
from decoupler.mt._aucell import aucell
from decoupler.mt._gsea import gsea
from decoupler.mt._gsva import gsva
from decoupler.mt._mdt import mdt
from decoupler.mt._mlm import mlm
from decoupler.mt._ora import ora
from decoupler.mt._udt import udt
from decoupler.mt._ulm import ulm
from decoupler.mt._viper import viper
from decoupler.mt._waggr import waggr
from decoupler.mt._zscore import zscore

_methods = [
    aucell,
    gsea,
    gsva,
    mdt,
    mlm,
    ora,
    udt,
    ulm,
    viper,
    waggr,
    zscore,
]



from itertools import product

import pandas as pd
import numpy as np

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import _download


def show_organisms(
) -> list:
    """
    Shows available organisms to translate to with ``decoupler.op.translate_net``.

    Returns
    -------
    List of available organisms.
    """
    valid_orgs = [
        'anole_lizard',
        'c.elegans',
        'cat',
        'cattle',
        'chicken',
        'chimpanzee',
        'dog',
        'fruitfly',
        'horse',
        'macaque',
        'mouse',
        'opossum',
        'pig',
        'platypus',
        'rat',
        's.cerevisiae',
        's.pombe',
        'xenopus',
        'zebrafish'
    ]
    return valid_orgs


def _replace_subunits(
    lst: list,
    my_dict: dict,
    one_to_many: int,
):
    result = []
    for x in lst:
        if x in my_dict:
            value = my_dict[x]
            if not isinstance(value, list):
                value = [value]
            if len(value) > one_to_many:
                result.append(np.nan)
            else:
                result.append(value)
        else:
            result.append(np.nan)
    return result


def _generate_orthologs(
    resource,
    column,
    map_dict,
    one_to_many
):
    df = resource[[column]].drop_duplicates().set_index(column)
    df["subunits"] = df.index.str.split("_")
    df["subunits"] = df["subunits"].apply(_replace_subunits, args=(map_dict, one_to_many, ))
    df = df["subunits"].explode().reset_index()
    grouped = df.groupby(column).filter(lambda x: x["subunits"].notna().all()).groupby(column)
    # Generate all possible subunit combinations within each group
    complexes = []
    for name, group in grouped:
        if group["subunits"].isnull().all():
            continue
        subunit_lists = [list(x) for x in group["subunits"]]
        complex_combinations = list(product(*subunit_lists))
        for complex in complex_combinations:
            complexes.append((name, "_".join(complex)))
    # Create output DataFrame
    col_names = ["orthology_source", "orthology_target"]
    result = pd.DataFrame(complexes, columns=col_names).set_index("orthology_source")
    return result


def _translate(
    resource: pd.DataFrame,
    map_dict: dict,
    column: str,
    one_to_many: int,
):  
    map_data = _generate_orthologs(resource, column, map_dict, one_to_many)
    resource = resource.merge(map_data, left_on=column, right_index=True, how="left")
    resource[column] = resource["orthology_target"]
    resource = resource.drop(columns=["orthology_target"])
    resource = resource.dropna(subset=[column])
    return resource


@docs.dedent
def translate(
    net: pd.DataFrame,
    columns: str | list = ['source', 'target', 'genesymbol'],
    target_organism: str = 'mouse',
    min_evidence: int | float = 3,
    one_to_many: int | float = 5,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Translates gene symbols from human to a target organism using the HCOP database.
    Check which organisms are available with ``decoupler.op.show_organisms``.

    HCOP is a composite database combining data from various orthology resources.
    It provides a comprehensive set of orthologs among human, mouse, and rat, among many other species.

    If you use this function, please reference the original HCOP paper:
    - Yates, B., Gray, K.A., Jones, T.E. and Bruford, E.A., 2021. Updates to HCOP: the HGNC comparison of orthology
    predictions tool. Briefings in Bioinformatics, 22(6), p.bbab155.

    For more information, please visit the HCOP website: https://www.genenames.org/tools/hcop/,
    or alternatively check the bulk download FTP links page: https://ftp.ebi.ac.uk/pub/databases/genenames/hcop/

    Parameters
    ----------
    %(net)s
    columns
        Column or columns of ``net`` to translate.
    target_organism
        Organism to translate to.
    min_evidence
        Minimum number of evidences to keep the interaction, where evidence is the number of
        orthology resources supporting the interaction.
    one_to_many
        Maximum number of orthologs allowed per gene.

    Returns
    -------
    Translated net.
    """
    # Validate
    assert isinstance(net, pd.DataFrame), 'net must be a pd.DataFrame'
    assert isinstance(columns, (str, list)), 'columns must be str or list'
    if isinstance(columns, str):
        columns = [columns]
    columns = [c for c in columns if c in net.columns]
    assert columns, f'columns must be one of these: {net.columns}'
    assert isinstance(target_organism, str), 'target_organism must be str'
    valid_orgs = show_organisms()
    assert target_organism in valid_orgs, f'target_organism must be one of these: {valid_orgs}'
    assert isinstance(min_evidence, (int, float)) and min_evidence > 0, 'min_evidence must be numerical and > 0'
    assert isinstance(one_to_many, (int, float)) and one_to_many > 0, 'one_to_many must be numerical and > 0'
    _log(f'Translating to {target_organism}', level='info', verbose=verbose)
    # Handle missmatch lavels from ebi db
    target_col = f'{target_organism}_symbol'
    if target_organism == 'anole_lizard':
        target_col = 'anole lizard_symbol'
    elif target_organism == 'fruitfly':
        target_col = 'fruit fly_symbol'
    # Process orthologs
    url = f'https://ftp.ebi.ac.uk/pub/databases/genenames/hcop/human_{target_organism}_hcop_fifteen_column.txt.gz'
    map_df = _download(url, low_memory=False, compression='gzip', sep='\t', verbose=verbose)
    map_df = pd.read_csv(url, sep="\t", low_memory=False)
    map_df['evidence'] = map_df['support'].apply(lambda x: len(x.split(",")))
    map_df = map_df[map_df['evidence'] >= min_evidence]
    map_df = map_df[[f'human_symbol', target_col]]
    map_df = map_df.rename(columns={'human_symbol': 'source', target_col: 'target'})
    map_dict = map_df.groupby('source')['target'].apply(list).to_dict()
    for col in columns:
        col_unique = net[col].drop_duplicates()
        prop = col_unique.isin(map_dict).sum() / col_unique.size
        m = f'Successfully translated {prop:.1%} of the genes from column {col}.'
        _log(m, level='info', verbose=verbose)
        net = _translate(resource=net, map_dict=map_dict, column=col, one_to_many=one_to_many)
    net = net.reset_index(drop=True)
    _log('Translation finished', level='info', verbose=verbose)
    return net



import pandas as pd
import numpy as np
from matplotlib.colors import to_rgb
from matplotlib.figure import Figure
import seaborn as sns

from decoupler._docs import docs
from decoupler._Plotter import Plotter


@docs.dedent
def order(
    df: pd.DataFrame,
    mode: str = 'line',
    kw_order = dict(),
    **kwargs
) -> None | Figure:
    """
    Plot features along a continuous, ordered process such as pseudotime.

    Parameters
    ----------
    df
        Results of ``decoupler.pp.bin_order``.
    mode
        The type of plot to use, either "line" or "mat".
    kw_order
        Other keyword arguments are passed down to ``seaborn.lineplot`` or ``matplotlib.pyplot.imshow``,
        depending on ``mode`` used.
    %(plot)s
    """
    # Validate
    assert isinstance(df, pd.DataFrame), 'df must be pandas.DataFrame'
    assert isinstance(mode, str) and mode in ['line', 'mat'], \
    'mode must be str and either "line" or "mat"'
    assert isinstance(kw_order, dict), \
    'kw_order must be dict'
    # Process
    ymax = df['value'].max()
    xmin, xmax = df['order'].min(), df['order'].max()
    n_names = df['name'].unique().size
    # Add cbar if added
    has_cbar = False
    if np.isin(['label', 'color'], df.columns).all():
        colors = df[df['name'] == df.loc[0, 'name']]['color']
        colors = [[to_rgb(c) for c in colors]]
        has_cbar = True
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    if mode == 'line':
        if has_cbar:
            bp.ax.imshow(
                colors,
                aspect='auto',
                extent=[0, xmax, 1.05 * ymax, 1.2 * ymax],
                transform=bp.ax.transData,
                zorder=2
            )
            bp.ax.axhline(y=1.05 * ymax, c='black', lw=1)
        kw_order = kw_order.copy()
        kw_order.setdefault('palette', 'tab20')
        sns.lineplot(
            data=df,
            x='order',
            y='value',
            hue='name',
            ax=bp.ax,
            **kw_order
        )
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False)
    elif mode == 'mat':
        mat = (
            df
            .groupby(['name', 'order'], as_index=False)['value'].mean()
            .pivot(index='name', columns='order', values='value')
        )
        img = bp.ax.imshow(mat, extent=[0, xmax, 0, n_names], aspect='auto', **kw_order)
        if has_cbar:
            bp.ax.imshow(colors, aspect='auto', extent=[0, xmax, n_names, 1.1 * n_names], zorder=2)
            bp.ax.axhline(y=n_names, c='black', lw=1)
            bp.ax.set_ylim(0, 1.1 * n_names)
        bp.fig.colorbar(img, ax=bp.ax, shrink=0.5, label='Mean value', location='top')
        bp.ax.set_yticks(np.arange(n_names) + 0.5)
        bp.ax.set_yticklabels(np.flip(mat.index))
        bp.ax.grid(axis='y', visible=False)
        bp.ax.set_xlabel('order')
    bp.ax.set_xlim(0, 1)
    return bp._return()



import numpy as np
import pandas as pd
import pytest

import gseapy as gp
import decoupler as dc


def test_std():
    arr = np.array([0.1, -5.3, 3.8, 9.5, -0.4, 5.5])
    np_std = np.std(arr, ddof=1)
    dc_std = dc.mt._gsea._std(arr=arr, ddof=1)
    assert np_std == dc_std


def test_ridx():
    idx_a = dc.mt._gsea._ridx(times=5, nvar=10, seed=42)
    assert (~(np.diff(idx_a) == 1).all(axis=1)).all()
    idx_b = dc.mt._gsea._ridx(times=5, nvar=10, seed=2)
    assert (~(np.diff(idx_b) == 1).all(axis=1)).all()
    assert (~(idx_a == idx_b).all(axis=1)).all()


def test_func_gsea(
    mat,
    net,
    idxmat,
):
    times = 1000
    seed = 42
    X, obs, var = mat
    gene_sets = net.groupby('source')['target'].apply(lambda x: list(x)).to_dict()
    cnct, starts, offsets = idxmat
    res = gp.prerank(
        rnk=pd.DataFrame(X, index=obs, columns=var).T,
        gene_sets=gene_sets,
        permutation_num=times,
        permutation_type='gene_set',
        outdir=None,
        min_size=0,
        threads=4,
        seed=seed,
    ).res2d
    gp_es = res.pivot(index='Name', columns='Term', values='NES').astype(float)
    gp_pv = res.pivot(index='Name', columns='Term', values='FDR q-val').astype(float)
    dc_es, dc_pv = dc.mt._gsea._func_gsea(
        mat=X,
        cnct=cnct,
        starts=starts,
        offsets=offsets,
        times=times,
        seed=seed,
    )
    assert (gp_es - dc_es).abs().values.max() < 0.10



from docrep import DocstringProcessor


_plot = """\
ax
    An existing :class:`matplotlib.axes._axes.Axes` instance to plot on. If ``None``,
    a new figure and axes will be created.
figsize
    Size of the figure in inches as (width, height).
dpi
    Dots per inch for the figure resolution.
return_fig
    If ``True``, plotting methods should return the figure object instead of showing it.
save
    If set, path to save the plot automatically to a file.

Returns
-------
If ``return_fig=True``, returns :class:`matplotlib.figure.Figure` instance.
"""

_features = """\
features
    Column names of ``mat``."""

_net = """\
net
    Dataframe in long format. Must include ``source`` and ``target`` columns, and optionally a ``weight`` column."""

_tmin = """\
tmin
    Minimum number of targets per source. Sources with fewer targets will be removed."""

_bsize = """\
bsize
    For large datasets in sparse format, this parameter controls how many observations are processed at once.
    Increasing this value speeds up computation but uses more memory."""

_verbose = """\
verbose
    Whether to display progress messages and additional execution details."""

_data = """\
data
    AnnData instance, DataFrame or tuple of [matrix, samples, features]."""

_layer = """\
layer
    Layer key name of an ``anndata.AnnData`` instance."""

_raw = """\
raw
    Whether to use the ``.raw`` attribute of ``anndata.AnnData``."""

_empty = """\
empty
    Whether to remove empty observations (rows) or features (columns)."""

_adata = """\
adata
    Annotated data matrix with observations (rows) and features (columns)."""

_times = """\
times
    Number of random permutations to do."""

_seed = """\
seed
    Random seed to use."""

_inplace = """\
inplace
    Whether to perform the operation in the same object."""

_min_cells = """\
min_cells
    Minimum number of cells per sample."""

_min_counts = """\
min_counts
    Minimum number of counts per sample."""

_key = """\
key
    ``adata.obsm`` key to use."""

_cmap = """\
cmap
    Colormap to use."""

_min_prop_prop = """\
min_prop
    Minimum proportion of cells that express a gene in a sample."""

_min_smpls = """\
min_smpls
    Minimum number of samples with bigger or equal proportion of cells with expression than ``min_prop``."""

_group = """\
group
    Name of the ``adata.obs`` column to group by. If None, it assumes that all samples belong to one group."""

_lib_size = """\
lib_size
    Library size. If None, default to the sum of reads per sample."""

_min_count = """\
min_count
    Minimum count requiered per gene for at least some samples."""

_min_total_count = """\
min_total_count
    Minimum total count required per gene across all samples."""

_large_n = """\
large_n
    Number of samples per group that is considered to be "large"."""

_min_prop_expr = """\
min_prop
    Minimum proportion of samples in the smallest group that express the gene."""

_organism = """\
organism
    The organism of interest. By default human."""

_license = """\
license
    Which license to use, available options are: academic, commercial, or nonprofit.
    By default, is set to academic to retrieve all possible interactions.
    Users are expected to comply with license regulations according to their affiliation."""

_df = """\
df
    Result of ``decoupler.bm.benchmark``."""

_hue = """\
hue
    Grouping variable that will produce different colors."""

_palette = """\
palette
    Method for choosing the colors to use"""

_y = """\
y
    Grouping variable to plot on y axis."""

_order = """\
order
    The name of the column in ``adata.obs`` to consider for ordering."""

_label = """\
label
    The name of the column in ``adata.obs`` to consider for coloring the grouping. By default ``None``."""

_nbins = """\
nbins
    Number of bins to use."""

_vmin = """\
vmin
    The value representing the lower limit of the color scale."""

_vcenter = """\
vcenter
    The value representing the center of the color scale."""

_vmax = """\
vmax
    The value representing the upper limit of the color scale."""

_data_plot = """\
data
    DataFrame containing feature-level statistics. Feature names must be in ``df.index``."""

_top = """\
top
    Number of top sources to plot."""

_yestest = """\
Finally, the obtained :math:`p_{value}` are adjusted by Benjamini-Hochberg correction."""

_notest = """\
This method does not perform statistical testing on :math:`ES` and therefore does not return :math:`p_{value}`."""

_returns = """\
Returns
-------

Enrichment scores :math:`ES` and, if applicable, adjusted :math:`p_{value}` by Benjamini-Hochberg.
"""

_params = f"""\
Parameters
----------
{_data}
{_net}
{_tmin}
{_raw}
{_empty}
{_bsize}
{_verbose}"""

docs = DocstringProcessor(
    plot=_plot,
    net=_net,
    tmin=_tmin,
    bsize=_bsize,
    verbose=_verbose,
    features=_features,
    data=_data,
    layer=_layer,
    raw=_raw,
    empty=_empty,
    adata=_adata,
    times=_times,
    seed=_seed,
    inplace=_inplace,
    min_cells=_min_cells,
    min_counts=_min_counts,
    key=_key,
    cmap=_cmap,
    min_prop_prop=_min_prop_prop,
    min_smpls=_min_smpls,
    group=_group,
    lib_size=_lib_size,
    min_count=_min_count,
    min_total_count=_min_total_count,
    large_n=_large_n,
    min_prop_expr=_min_prop_expr,
    organism=_organism,
    license=_license,
    df=_df,
    hue=_hue,
    palette=_palette,
    y=_y,
    order=_order,
    label=_label,
    nbins=_nbins,
    vmin=_vmin,
    vcenter=_vcenter,
    vmax=_vmax,
    data_plot=_data_plot,
    top=_top,
    params=_params,
    yestest=_yestest,
    notest=_notest,
    returns=_returns,
)



from typing import Tuple
import math

import numpy as np
import scipy.stats as sts
import scipy.sparse as sps
from tqdm.auto import tqdm
import numba as nb

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method
from decoupler.pp.net import _getset
from decoupler.mt._gsea import _std


@nb.njit(cache=True)
def _erf(
    x: np.ndarray,
) -> np.ndarray:
    a1, a2, a3, a4, a5, a6 = 0.254829592, -0.284496736, 1.421413741, -1.453152027, 1.061405429, 0.3275911
    sign = np.sign(x)
    abs_x = np.abs(x)
    t = 1.0 / (1.0 + a6 * abs_x)
    y = 1.0 - (((((a5 * t + a4) * t + a3) * t + a2) * t + a1) * t * np.exp(-abs_x * abs_x))
    return sign * y


@nb.njit(cache=True)
def _norm_cdf(
    x: np.ndarray,
    mu: float = 0.0,
    sigma: float = 1.0,
) -> np.ndarray:
    e = _erf((x - mu) / (sigma * np.sqrt(2.0)))
    return (0.5 * (1.0 + e))


@nb.njit(cache=True)
def _poisson_pmf(
    k: float,
    lam: float,
) -> float:
    if k < 0 or lam < 0:
        return 0.0
    if k == 0:
        return np.exp(-lam)
    log_pmf = -lam + k * np.log(lam) - math.lgamma(k + 1)
    return np.exp(log_pmf)


@nb.njit(cache=True)
def _ppois(
    k: float,
    lam: float
) -> float:
    cdf_sum = 0.0
    for i in range(int(k) + 1):
        cdf_sum += _poisson_pmf(i, lam)
    if cdf_sum > 1:
        cdf_sum = 1.
    return cdf_sum


@nb.njit(cache=True)
def _init_cdfs(
) -> np.ndarray:
    pre_res = 10000
    max_pre = 10
    pre_cdf = _norm_cdf(np.arange(pre_res + 1) * max_pre / pre_res, 0, 1)
    return pre_cdf


@nb.njit(cache=True)
def _ecdf(arr):
    ecdf = np.searchsorted(np.sort(arr), arr, side='right') / len(arr)
    return ecdf


@nb.njit(parallel=True, cache=True)
def _mat_ecdf(
    mat: np.ndarray
) -> np.ndarray:
    D = np.zeros(mat.shape)
    for j in range(mat.shape[1]):
        D[:, j] = _ecdf(mat[:, j])
    return D
    

@nb.njit(cache=True)
def _col_d(
    x: np.ndarray,
    gauss: bool,
    pre_cdf: np.ndarray
) -> np.ndarray:
    size = x.shape[0]
    if gauss:
        bw = _std(x, 1) / 4.0
    else:
        bw = 0.5
    col = np.zeros(size)
    for j in range(size):
        left_tail = 0.0
        for i in range(size):
            if gauss:
                diff = (x[j] - x[i]) / bw
                if diff < -10:
                    left_tail += 0.0
                elif diff > 10:
                    left_tail += 1.0
                else:
                    cdf_val = pre_cdf[int(np.abs(diff) / 10 * 10000)]
                    if diff < 0:
                        left_tail += 1.0 - cdf_val
                    else:
                        left_tail += cdf_val
            else:
                left_tail += _ppois(x[j], x[i] + bw)
        left_tail = left_tail / size
        col[j] = -1.0 * np.log((1.0 - left_tail) / left_tail)
    return col


@nb.njit(parallel=True, cache=True)
def _mat_d(
    mat: np.ndarray,
    gauss: bool
) -> np.ndarray:
    pre_cdf = _init_cdfs()
    D = np.zeros(mat.shape)
    for j in nb.prange(mat.shape[1]):
        D[:, j] = _col_d(mat[:, j], gauss, pre_cdf)
    return D


def _density(
    mat: np.ndarray,
    kcdf: str | None,
) -> np.ndarray:
    assert (isinstance(kcdf, str) and kcdf in ['gaussian', 'poisson']) or kcdf is None, \
    'kcdf must be gaussian, poisson or None'
    if kcdf == 'gaussian':
        mat = _mat_d(mat, gauss=True)
    elif kcdf == 'poisson':
        assert mat.sum().is_integer(), \
        f'when kcdf={kcdf} input data must be integers (e.g. 3, 4, etc.), not decimal values (e.g. 3.5, 4.9, etc.)'
        mat = _mat_d(mat, gauss=False)
    elif kcdf is None:
        mat = _mat_ecdf(mat)
    return mat


@nb.njit(cache=True)
def _rankdata(values):
    n = len(values)
    ranks = np.empty(n, dtype=np.int_)
    indices = np.arange(n)
    sorted_indices = np.empty(n, dtype=np.int_)
    sorted_values = np.empty(n, dtype=values.dtype)
    for i in range(n):
        sorted_indices[i] = indices[i]
        sorted_values[i] = values[i]
    for i in range(n):
        for j in range(i + 1, n):
            vi, vj = sorted_values[i], sorted_values[j]
            ii, ij = sorted_indices[i], sorted_indices[j]
            if (vj < vi) or (vj == vi and ij > ii):
                sorted_values[i], sorted_values[j] = vj, vi
                sorted_indices[i], sorted_indices[j] = ij, ii
    for rank, idx in enumerate(sorted_indices, 1):
        ranks[idx] = rank
    return ranks


@nb.njit(cache=True)
def _dos_srs(r):
    mask = (r == 0)
    p = len(r)
    r_dense = r.astype(np.int_).copy()
    if mask.any():
        nzs = mask.sum()
        r_dense[~mask] += nzs
        cnt = 1
        for i in range(p):
            if mask[i]:
                r_dense[i] = cnt
                cnt += 1
    dos = p - r_dense + 1
    srs = np.empty(p)
    if mask.any():
        r_mod = r.copy()
        for i in range(p):
            if mask[i]:
                r_mod[i] = 1
            else:
                r_mod[i] += 1
        max_r = np.max(r_mod)
        half_max = max_r / 2
        for i in range(p):
            srs[i] = abs(half_max - r_mod[i])
    else:
        half_p = p / 2
        for i in range(p):
            srs[i] = abs(half_p - r_dense[i])
    return dos, srs


@nb.njit(parallel=True, cache=True)
def _rankmat(
    mat: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    n_rows, n_cols = mat.shape
    dos_mat = np.zeros((n_rows, n_cols), dtype=np.int_)
    srs_mat = np.zeros((n_rows, n_cols), dtype=np.int_)
    for i in nb.prange(n_rows):
        r = _rankdata(mat[i, :])
        dos_mat[i, :], srs_mat[i, :] = _dos_srs(r)
    return dos_mat, srs_mat


@nb.njit(cache=True)
def _rnd_walk(
    gsetidx: np.ndarray,
    k: int,
    decordstat: np.ndarray,
    symrnkstat: np.ndarray,
    n: int,
    tau: int | float,
) -> Tuple[float, int]:
    gsetrnk = np.empty(k, dtype=np.int_)
    for i in range(k):
        gsetrnk[i] = decordstat[gsetidx[i] - 1]
    stepcdfingeneset = np.zeros(n)
    stepcdfoutgeneset = np.ones(n, dtype=np.int_)
    for i in range(k):
        idx = gsetrnk[i] - 1
        if tau == 1.0:
            stepcdfingeneset[idx] = symrnkstat[gsetidx[i] - 1]
        else:
            stepcdfingeneset[idx] = symrnkstat[gsetidx[i] - 1] ** tau
        stepcdfoutgeneset[idx] = 0
    # cumulative sums
    for i in range(1, n):
        stepcdfingeneset[i] += stepcdfingeneset[i - 1]
        stepcdfoutgeneset[i] += stepcdfoutgeneset[i - 1]
    walkstatpos = -np.inf
    walkstatneg = np.inf
    if stepcdfingeneset[n - 1] > 0 and stepcdfoutgeneset[n - 1] > 0:
        walkstatpos = 0.0
        walkstatneg = 0.0
        for i in range(n):
            wlkstat = (stepcdfingeneset[i] / stepcdfingeneset[n - 1]) - \
                      (stepcdfoutgeneset[i] / stepcdfoutgeneset[n - 1])
            if wlkstat > walkstatpos:
                walkstatpos = wlkstat
            if wlkstat < walkstatneg:
                walkstatneg = wlkstat
    return walkstatpos, walkstatneg


@nb.njit(cache=True)
def _score_geneset(
    gsetidx: np.ndarray,
    generanking: np.ndarray,
    rankstat: np.ndarray,
    maxdiff: bool,
    absrnk: bool,
    tau: int | float,
) -> float:
    n = len(generanking)
    k = len(gsetidx)
    walkstatpos, walkstatneg = _rnd_walk(
        gsetidx=gsetidx, k=k, decordstat=generanking, symrnkstat=rankstat, n=n, tau=tau
    )
    if maxdiff:
        if absrnk:
            es = walkstatpos - walkstatneg
        else:
            es = walkstatpos + walkstatneg
    else:
        es = walkstatpos if abs(walkstatpos) > abs(walkstatneg) else walkstatneg
    return es


@nb.njit(parallel=True, cache=True)
def _ks_fset(
    dos: np.ndarray,
    srs: np.ndarray,
    fset: np.ndarray,
    maxdiff: bool,
    absrnk: bool,
    tau: int | float,
) -> np.ndarray:
    n_samples, n_genes = dos.shape
    res = np.zeros(n_samples)
    for i in nb.prange(n_samples):
        generanking = dos[i]
        rankstat = srs[i]
        genesetsrankidx = fset
        res[i] = _score_geneset(genesetsrankidx, generanking, rankstat, maxdiff, absrnk, tau)
    return res


@docs.dedent
def _func_gsva(
    mat: np.ndarray,
    cnct: np.ndarray,
    starts: np.ndarray,
    offsets: np.ndarray,
    kcdf: str | None = 'gaussian',
    maxdiff: bool = True,
    absrnk: bool = False,
    tau: int | float = 1,
    verbose: bool = False,
) -> Tuple[np.ndarray, None]:
    r"""
    Gene Set Variation Analysis (GSVA) :cite:`gsva`.

    Each feature is first transformed and smoothed using a kernel density estimation method:
    
    - Gaussian
    - Poisson
    - Empirical cumulative distribution function

    Features are then ranked based on a continuous metric (e.g., expression value, score, or correlation).
    
    Then, a score for each feature in a set is computed by walking down the ranked list,
    increasing a running-sum statistic when a feature belongs to the set and decreasing it otherwise.

    .. math::
    
       \delta(F, i) = 
       \begin{cases}
       \frac{|r_i|}{\sum\limits_{j \in F} |r_j|} & \text{if feature } i \in F \\
       -\frac{1}{l} & \text{if feature } i \notin F
       \end{cases}

    Where:

    - :math:`F` is a feature set
    - :math:`r` is the ranking of the feature statistics in descending order
    - :math:`r_i` is the value for feature :math:`i`
    - :math:`r_j` is the value for feature :math:`j` in :math:`F`
    - :math:`k` is the number of features in :math:`F`
    - :math:`N` is the total number of features in :math:`r`
    - :math:`l=N-k` is the number of features not in :math:`F` but present in :math:`r`

    For each feature, the function :math:`\delta(F,i)` is applied and stored as a sequence :math:`L`.

    .. math::

        L = \delta(F, i)\text{ for i} = \text{1, 2, ... , N}

    The enrichment score :math:`ES` is computed as the sum of the maximum positive and maximum negative deviations
    of the running-sum statistic from zero.

    .. math::

        ES = \max_{1 \leq i \leq N} L_i + \min_{1 \leq i \leq N} L_i

    %(notest)s

    %(params)s
    kcdf
        Which kernel to use during the non-parametric estimation of the cumulative distribution function.
        Options are gaussian, poisson or None.
    mx_diff
        Changes how the enrichment statistic (ES) is calculated.
        If ``True`` (default), ES is calculated as the difference between the maximum positive and
        negative random walk deviations.
        If ``False``, ES is calculated as the maximum positive to 0.
    abs_rnk : bool
        Used when ``mx_diff=True``. If ``False`` (default), the enrichment statistic (ES) is calculated taking the magnitude
        difference between the largest positive and negative random walk deviations.
        If ``True``, feature sets with features enriched on either extreme (high or low)
        will be regarded as 'highly' activated.

    %(returns)s
    """
    if isinstance(mat, sps.csr_matrix):
        m = f'gsva - Converting sparse matrix to dense format before density transformation'
        _log(m, level='info', verbose=verbose)
        mat = mat.toarray()
    m = f'gsva - computing density with kcdf={kcdf}'
    _log(m, level='info', verbose=verbose)
    # Compute density
    if mat.shape[0] > 1:
        mat = _density(mat, kcdf=kcdf)
    dos, srs = _rankmat(mat)
    # Compute GSVA
    nsrc = starts.size
    m = f'gsva - calculating {nsrc} scores with maxdiff={maxdiff}, absrnk={absrnk}'
    _log(m, level='info', verbose=verbose)
    es = np.zeros((dos.shape[0], nsrc))
    for j in tqdm(range(nsrc), disable=not verbose):
        fset = (_getset(cnct, starts, offsets, j) + 1).astype(int)
        es[:, j] = _ks_fset(dos=dos, srs=srs, fset=fset, maxdiff=maxdiff, absrnk=absrnk, tau=tau)
    return es, None


_gsva = MethodMeta(
    name='gsva',
    desc='Gene Set Variation Analysis (GSVA)',
    func=_func_gsva,
    stype='numerical',
    adj=False,
    weight=False,
    test=False,
    limits=(-1, +1),
    reference='https://doi.org/10.1186/1471-2105-14-7',
)
gsva = Method(_method=_gsva)



import numpy as np
import pandas as pd

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import URL_INT, _download
from decoupler.op._translate import translate
from decoupler.op._dtype import _infer_dtypes


@docs.dedent
def dorothea(
    organism: str = 'human',
    levels: str | list = ['A', 'B', 'C'],
    dict_weights: dict | None = None,
    license: str = 'academic',
    verbose: bool = False,
) -> pd.DataFrame:
    """
    DoRothEA gene regulatory network :cite:p:`dorothea`.

    Wrapper to access DoRothEA gene regulatory network. DoRothEA is a
    comprehensive resource containing a curated collection of transcription
    factors (TFs) and their target genes. Each interaction is weighted by its
    mode of regulation (either positive or negative) and by its confidence
    level.

    Parameters
    ----------
    %(organism)s
    levels
        List of confidence levels to return. Goes from A to D, A being the
        most confident and D being the less.
    dict_weights
        Dictionary of values to divide the mode of regulation (-1 or 1),
        one for each confidence level. Bigger values will generate weights
        close to zero.
    %(license)s
    %(verbose)s

    Returns
    -------
    Dataframe in long format containing target genes for each TF with their associated weights and confidence level.
    """
    assert isinstance(levels, (str, list)), 'levels must be str or list'
    if isinstance(levels, str):
        levels = [levels]
    assert all(l in {'A', 'B', 'C', 'D'} for l in levels), 'levels can only contain any of these values: A, B, C, and/or D'
    assert isinstance(dict_weights, dict) or dict_weights is None, 'dict_weights must be dict or None'
    if dict_weights:
        assert all(k in levels for k in dict_weights), f'dict_weights keys must be in levels={levels}'
        weights = dict_weights
    else:
        weights = {'A': 1, 'B': 2, 'C': 3, 'D': 4}
        weights = {k: weights[k] for k in weights if k in levels}
    # Read
    str_levels = ','.join(levels)
    url_ext = f'datasets=dorothea&dorothea_levels={str_levels}&fields=dorothea_level&license={license}'
    url = URL_INT + url_ext
    m = f'dorothea - Accessing DoRothEA (levels {str_levels}) with {license} license and weights={weights}'
    _log(m, level='info', verbose=verbose)
    do = _download(url, sep='\t', verbose=verbose)
    # Filter extra columns
    do = do[[
        'source_genesymbol', 'target_genesymbol',
        'is_stimulation', 'is_inhibition',
        'consensus_direction', 'consensus_stimulation',
        'consensus_inhibition', 'dorothea_level',
    ]]
    # Remove duplicates
    do = do[~do.duplicated(['source_genesymbol', 'dorothea_level', 'target_genesymbol'])]
    # Assign top level if more than 2
    do['dorothea_level'] = [lvl.split(';')[0] for lvl in do['dorothea_level']]
    # Assign mode of regulation
    mor = []
    for i in do.itertuples():
        if i.is_stimulation and i.is_inhibition:
            if i.consensus_stimulation:
                mor.append(1)
            else:
                mor.append(-1)
        elif i.is_stimulation:
            mor.append(1)
        elif i.is_inhibition:
            mor.append(-1)
        else:
            mor.append(1)
    do['mor'] = mor
    # Compute weight based on confidence: mor/confidence
    do['weight'] = [i.mor / weights[i.dorothea_level] for i in do.itertuples()]
    # Format
    do = (
        do
        .rename(columns={'source_genesymbol': 'source', 'target_genesymbol': 'target', 'dorothea_level': 'confidence'})
        [['source', 'target', 'weight', 'confidence']]
        .sort_values('confidence')
    )
    do = do[do['confidence'].isin(levels)].reset_index(drop=True)
    do = _infer_dtypes(do)
    if organism != 'human':
        do = translate(do, columns=['source', 'target'], target_organism=organism, verbose=verbose)
    do = do.drop_duplicates(['source', 'target']).reset_index(drop=True)
    return do



from typing import Tuple

import numpy as np
import pandas as pd
from anndata import AnnData

from decoupler._docs import docs
from decoupler._log import _log


def _fillval(
    arr: np.ndarray,
    nvar: int,
    val: float,
) -> np.ndarray:
    # Compute how many zeros to add
    n_missing = nvar - arr.size
    # Pad with zeros if needed
    if n_missing > 0:
        arr = np.pad(arr, (0, n_missing), constant_values=val)
    return arr


@docs.dedent
def toy(
    nobs: int = 30,
    nvar: int = 20,
    bval: int | float = 2.,
    pstime: bool = False,
    seed: int = 42,
    verbose: bool = False,
) -> Tuple[AnnData, pd.DataFrame]:
    """
    Generate a toy adata and net for testing.

    Parameters
    ----------
    nobs
        Number of samples to generate.
    nvar
        Number of features to generate.
    bvar
        Background value to set features not associated to any source.
    pstime
        Whether to add simulated pseudotime.
    %(seed)s
    %(verbose)s

    Returns
    -------
    AnnData and net examples.
    """
    # Validate
    assert isinstance(nobs, (int, float)) and nobs >= 2, \
    'nobs must be numeric and >= 2'
    assert isinstance(nvar, (int, float)) and nvar >= 12, \
    'nvar must be numeric and >= 12'
    assert isinstance(bval, (int, float)), \
    'bval must be numeric'
    # Network model
    net = pd.DataFrame([
        ['T1', 'G01', 1], ['T1', 'G02', 1], ['T1', 'G03', 0.7],
        ['T2', 'G04', 1], ['T2', 'G06', -0.5], ['T2', 'G07', -3], ['T2', 'G08', -1],
        ['T3', 'G06', 1], ['T3', 'G07', 0.5], ['T3', 'G08', 1],
        ['T4', 'G05', 1.9], ['T4', 'G10', -1.5], ['T4', 'G11', -2], ['T4', 'G09', 3.1],
        ['T5', 'G09', 0.7], ['T5', 'G10', 1.1], ['T5', 'G11', 0.1],
    ], columns=['source', 'target', 'weight'])
    # Simulate two population of samples with different molecular values
    rng = np.random.default_rng(seed=seed)
    n = int(nobs / 2)
    res = nobs % 2
    row_a = np.array([8, 8, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0])
    row_b = np.array([0, 0, 0, 0, 8, 8, 8, 8, 0, 0, 0, 0])
    if nvar > 12:
        m = f'toy - adding background gene expresison for extra vars bval={bval}'
        _log(m, level='info', verbose=verbose)
        row_a = _fillval(row_a, nvar=nvar, val=bval)
        row_b = _fillval(row_b, nvar=nvar, val=bval)
    m = f'toy - adding random noise to .X with seed={seed}'
    _log(m, level='info', verbose=verbose)
    row_a = [row_a + np.abs(rng.normal(size=nvar)) for _ in range(n)]
    row_b = [row_b + np.abs(rng.normal(size=nvar)) for _ in range(n + res)]
    adata = np.vstack([row_a, row_b])
    features = ['G{:02d}'.format(i + 1) for i in range(nvar)]
    samples = ['C{:02d}'.format(i + 1) for i in range(nobs)]
    adata = pd.DataFrame(adata, index=samples, columns=features)
    adata = AnnData(adata)
    adata.obs['group'] = (['A'] * len(row_a)) + (['B'] * len(row_b))
    adata.obs['group'] = adata.obs['group'].astype('category')
    adata.obs['sample'] = rng.choice(['S01', 'S02', 'S03'], size=adata.n_obs, replace=True)
    adata.obs['sample'] = adata.obs['sample'].astype('category')
    if pstime:
        m = f'toy - Adding simulated pseudotime'
        _log(m, level='info', verbose=verbose)
        pst = np.arange(adata.n_obs)
        pst = pst / pst.max()
        adata.X[:, :8] = adata.X[:, :8] + (adata.X[:, :8] * pst.reshape(-1, 1))
        adata.obs['pstime'] = pst
    m = f'toy - generated AnnData with shape={adata.shape}'
    _log(m, level='info', verbose=verbose)
    return adata, net


@docs.dedent
def toy_bench(
    shuffle_r: float = 0.25,
    seed: int = 42,
    verbose: bool = False,
    **kwargs
):
    """
    Generate a toy adata and net for testing the benchmark pipeline.

    Parameters
    ----------
    shuffle_r
        Percentage of the ground truth to randomize.
    %(seed)s
    %(verbose)s
    kwargs
        All other keyword arguments are passed to ``decoupler.ds.toy``.

    Returns
    -------
    AnnData and net examples.
    """
    # Validate
    assert isinstance(shuffle_r, (int, float)) and 0.0 <= shuffle_r <= 1.0, \
    'shuffle_r must be numeric and between 0 and 1'
    # Get toy data
    adata, net = toy(**kwargs)
    # Add grount truth
    adata.obs['source'] = [['T1', 'T2'] if g == 'A' else ['T3', 'T4'] for g in adata.obs['group']]
    adata.obs['class'] = np.tile(['CA', 'CB'], int(np.ceil(adata.obs_names.size / 2)))[:adata.obs_names.size]
    adata.obs['class'] = adata.obs['class'].astype('category')
    adata.obs['type_p'] = 1.
    # Shuffle a percentage of the samples
    idxs = np.arange(adata.obs_names.size)
    n_shuffle = int(np.ceil(idxs.size * shuffle_r))
    if n_shuffle > 0:
        rng = np.random.default_rng(seed=seed)
        m = f'Shuffling {n_shuffle} observations ({shuffle_r * 100:.2f}%).'
        _log(m, level='info', verbose=verbose)
        idxs = rng.choice(idxs, n_shuffle, replace=False)
        r_idxs = rng.choice(idxs, idxs.size, replace=False)
        adata.X[r_idxs, :] = adata.X[idxs, :]
    return adata, net



import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.fixture
def df():
    df = pd.DataFrame(
        data=[
            [1, -2, 3, -4],
            [5, -6, 7, -8],
        ],
        index=['C1', 'C2'],
        columns=[f'TF{i}' for i in range(4)]
    )
    return df


@pytest.mark.parametrize(
    'name,top,vertical',
    [
        ['C1', 2, True],
        ['C2', 10, False],
    ]
)
def test_barplot(
    df,
    name,
    top,
    vertical,
):
    fig = dc.pl.barplot(data=df, name=name, top=top, vertical=vertical, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



import numpy as np
import scipy.stats as sts
import pytest

import decoupler as dc


def test_cov(
    mat,
    adjmat,
):
    X, obs, var = mat
    dc_cov = dc.mt._ulm._cov(A=adjmat, b=X.T)
    nsrcs = adjmat.shape[1]
    np_cov = np.cov(m=adjmat, y=X.T, rowvar=False)[:nsrcs, nsrcs:].T
    assert np.allclose(np_cov, dc_cov)


def test_cor(
    mat,
    adjmat,
):
    X, obs, var = mat
    dc_cor = dc.mt._ulm._cor(adjmat, X.T)
    nsrcs = adjmat.shape[1]
    np_cor = np.corrcoef(adjmat, X.T, rowvar=False)[:nsrcs, nsrcs:].T
    assert np.allclose(dc_cor, np_cor)
    assert np.all((dc_cor <= 1) * (dc_cor >= -1))


def test_tval():
    t = dc.mt._ulm._tval(r=0.4, df=28)
    assert np.allclose(2.30940108, t)
    t = dc.mt._ulm._tval(r=0.99, df=3)
    assert np.allclose(12.15540081, t)
    t = dc.mt._ulm._tval(r=-0.05, df=99)
    assert np.allclose(-0.49811675, t)


def test_func_ulm(
    mat,
    adjmat,
):
    X, obs, var = mat
    dc_es, dc_pv = dc.mt._ulm._func_ulm(mat=X, adj=adjmat)
    st_es, st_pv = np.zeros(dc_es.shape), np.zeros(dc_pv.shape)
    for i in range(st_es.shape[0]):
        for j in range(st_es.shape[1]):
            x = X[i, :]
            y = adjmat[:, j]
            slope, _, _, st_pv[i, j], std_err = sts.linregress(x, y)
            st_es[i, j] = slope / std_err
    assert np.allclose(dc_es, st_es)
    assert np.allclose(dc_pv, st_pv)
    
    


from typing import Tuple

import pandas as pd
import scipy.stats as sts
from anndata import AnnData

from decoupler._docs import docs


def _input_rank_obsm(
    adata: AnnData,
    key: str,
) -> Tuple[pd.DataFrame, list, list]:
    # Validate
    assert isinstance(adata, AnnData), 'adata must be anndata.AnnData'
    assert key in adata.obsm, f'key={key} must be in adata.obsm'
    # Process
    name_col = (
        key
        .replace('X_', '')
        .replace('pca', 'PC')
        .replace('mofa', 'Factor')
        .replace('umap', 'UMAP')
    )
    df = adata.obsm[key]
    if isinstance(df, pd.DataFrame):
        y_vars = df.std(ddof=1, axis=0).sort_values(ascending=False).index
        df = df.loc[:, y_vars].values
    else:
        ncol = df.shape[1]
        digits = len(str(ncol))
        y_vars = [f"{name_col}{str(i).zfill(digits)}" for i in range(1, ncol + 1)]
    df = pd.DataFrame(
        data=df,
        index=adata.obs_names,
        columns=y_vars
    )
    x_vars = adata.obs.columns
    # Merge
    df = pd.merge(df, adata.obs, left_index=True, right_index=True)
    return df, x_vars, y_vars


@docs.dedent
def rankby_obsm(
    adata: AnnData,
    key: str,
    uns_key: str | None = 'rank_obsm',
) -> None | pd.DataFrame:
    """
    Ranks features in ``adata.obsm`` by the significance of their association with metadata in ``adata.obs``.

    For categorical variables it uses ANOVA, for continous Spearman's correlation.

    The obtained p-values are corrected by Benjamini-Hochberg.

    Parameters
    ----------
    %(adata)s
    %(key)s
    uns_key
        ``adata.uns`` key to store the results.

    Returns
    -------
    If ``uns_key=False``, a pandas.DataFrame with the resulting statistics.
    """
    assert isinstance(uns_key, str) or uns_key is None, \
    'uns_key must be str or None'
    # Extract
    df, x_vars, y_vars = _input_rank_obsm(adata=adata, key=key)
    # Test
    res = []
    for x_var in x_vars:
        for y_var in y_vars:
            if pd.api.types.is_numeric_dtype(df[x_var]):
                # Correlation
                x = df[x_var].values.ravel()
                y = df[y_var].values.ravel()
                stat, pval = sts.spearmanr(x, y)
            else:
                # ANOVA
                x = [group[y_var].dropna().values for _, group in df.groupby(x_var, observed=True)]
                # At least n=2 per group else skip
                if all(len(g) >= 2 for g in x):
                    stat, pval = sts.f_oneway(*x)
                else:
                    stat, pval = None, 1.
            row = [y_var, x_var, stat, pval]
            res.append(row)
    res = pd.DataFrame(res, columns=['obsm', 'obs', 'stat', 'pval'])
    res['padj'] = sts.false_discovery_control(res['pval'])
    # Rank
    res = res.sort_values('padj').reset_index(drop=True)
    # Add obsm key
    res.key = key
    # Save or return
    if uns_key:
        adata.uns[uns_key] = res
    else:
        return res



import pandas as pd
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'groupby,reference,method',
    [
        ['group', 'rest', 'wilcoxon'],
        ['group', 'A', 't-test'],
        ['group', 'A', 't-test_overestim_var'],
        ['sample', 'rest', 't-test_overestim_var'],
        ['sample', 'S01', 't-test_overestim_var'],
        ['sample', ['S01'], 't-test_overestim_var'],
        ['sample', ['S01', 'S02'], 't-test_overestim_var'],
    ]
)
def test_rankby_group(
    adata,
    groupby,
    reference,
    method,
):
    df = dc.tl.rankby_group(adata=adata, groupby=groupby, reference=reference, method=method)
    assert isinstance(df, pd.DataFrame)
    cols_cat = {'group', 'reference', 'name'}
    cols_num = {'stat', 'meanchange', 'pval', 'padj'}
    cols = cols_cat | cols_num
    assert cols.issubset(set(df.columns))
    for col in cols_cat:
        assert isinstance(df[col].dtype, pd.CategoricalDtype)
    for col in cols_num:
        assert pd.api.types.is_numeric_dtype(df[col])
    assert set(df['group'].cat.categories).issubset(set(adata.obs[groupby].cat.categories))
    assert ((0. <= df['padj']) & (df['padj'] <= 1.)).all()



import logging
import tempfile

import pandas as pd
import numpy as np
import pytest

import decoupler as dc


def test_read_gmt():
    gmt = 'S1\tlink\tG1\tG2\tG3\nS2\tlink\tG2\tG3\tG4\tG5\tG6'
    with tempfile.NamedTemporaryFile(mode='w+', suffix='.txt', delete=True) as tmp:
        tmp.write(gmt)
        tmp.flush()
        gmt = dc.pp.read_gmt(tmp.name)
    assert isinstance(gmt, pd.DataFrame)
    assert {'source', 'target'}.issubset(gmt.columns)


@pytest.mark.parametrize('verbose', [True, False])
def test_validate_net(
    net,
    verbose,
    caplog,
):
    with caplog.at_level(logging.WARNING):
        vnet = dc.pp.net._validate_net(net=net, verbose=verbose)
    assert caplog.text == ''
    assert net.shape == vnet.shape
    assert 'source' in vnet.columns
    assert 'target' in vnet.columns
    assert 'weight' in vnet.columns

    net.drop(columns=['weight'], inplace=True)
    assert 'weight' not in net.columns
    with caplog.at_level(logging.WARNING):
        vnet = dc.pp.net._validate_net(net=net, verbose=verbose)
    if verbose:
        assert len(caplog.text) > 0
    assert 'weight' in vnet.columns

    enet = net.rename(columns={'source': 'tf'})
    with pytest.raises(AssertionError):
        dc.pp.net._validate_net(enet)


@pytest.mark.parametrize(
    'tmin,nrow_equal,raise_err',
    [
        [0, True, False],
        [3, True, False],
        [4, False, False],
        [5, False, True],
    ]
)
def test_prune(
    net,
    tmin,
    nrow_equal,
    raise_err,
):
    features = ['G{:02d}'.format(i + 1) for i in range(20)]
    if raise_err:
        with pytest.raises(AssertionError):
            dc.pp.net.prune(features=features, net=net, tmin=tmin)
    else:
        pnet = dc.pp.net.prune(features=features, net=net, tmin=tmin)
        if nrow_equal:
            assert pnet.shape[0] == net.shape[0]
        else:
            assert pnet.shape[0] < net.shape[0]


def test_adj(
    net,
):
    sources, targets, adj = dc.pp.net._adj(net=net)
    sources, targets = list(sources), list(targets)
    net = net.set_index(['source', 'target'])
    for s in sources:
        j = sources.index(s)
        snet = net.loc[s]
        for t in snet.index:
            i = targets.index(t)
            w_adj = adj[i, j]
            w_net = snet.loc[t]['weight']
            print(s, t, i, j, w_net, w_adj)
            assert w_net == w_adj
    assert (len(targets), len(sources)) == adj.shape
    assert adj.dtype == float


def test_order(
    net
):
    sources, targets, adjmat = dc.pp.net._adj(net=net)

    rtargets = targets[::-1]
    radjmat = adjmat[::-1]
    oadjmat = dc.pp.net._order(features=targets, targets=rtargets, adjmat=radjmat)
    assert (adjmat == oadjmat).all()
    
    mfeatures = list(targets) + ['x', 'y', 'z']
    madjmat = dc.pp.net._order(features=mfeatures, targets=targets, adjmat=adjmat)
    assert np.all(madjmat == 0, axis=1).sum() == 3
    
    lfeatures = targets[:5]
    assert lfeatures.size < targets.size
    ladjmat = dc.pp.net._order(features=lfeatures, targets=targets, adjmat=adjmat)
    assert ladjmat.shape[0] < adjmat.shape[0]
    assert np.all(ladjmat == 0, axis=1).sum() == 0
    

@pytest.mark.parametrize('verbose', [True, False])
def test_adjmat(
    adata,
    net,
    unwnet,
    verbose,
    caplog,
):
    features = adata.var_names
    with caplog.at_level(logging.INFO):
        sources, targets, adjmat = dc.pp.adjmat(
            features=features, net=net, verbose=verbose
        )
    adjmat = adjmat.ravel()
    non_zero_adjmat = adjmat[adjmat != 0.]
    assert not all(non_zero_adjmat == 1.)
    if verbose:
        assert len(caplog.text) > 0
    else:
        assert caplog.text == ''
    unwnet = dc.pp.net._validate_net(net=unwnet)
    sources, targets, adjmat = dc.pp.adjmat(
        features=features, net=unwnet
    )
    adjmat = adjmat.ravel()
    non_zero_adjmat = adjmat[adjmat != 0.]
    assert all(non_zero_adjmat == 1.)
    

@pytest.mark.parametrize('verbose', [True, False])
def test_idxmat(
    adata,
    net,
    verbose,
    caplog,
):
    features = adata.var_names
    with caplog.at_level(logging.INFO):
        sources, cnct, starts, offsets = dc.pp.idxmat(features=features, net=net, verbose=verbose)
    if verbose:
        assert len(caplog.text) > 0
    else:
        assert caplog.text == ''
    assert sources.size == starts.size
    assert starts.size == offsets.size
    assert (net.groupby('source')['target'].size().loc[sources] == offsets).all()
    assert cnct.size == offsets.sum()


@pytest.mark.parametrize('j', [0, 1, 2, 3])
def test_getset(
    idxmat,
    j,
):
    cnct, starts, offsets = idxmat
    fset = dc.pp.net._getset(cnct=cnct, starts=starts, offsets=offsets, j=j)
    assert isinstance(fset, np.ndarray)
    assert fset.size == offsets[j]


@pytest.mark.parametrize('seed', [1, 2, 3])
def test_shuffle_net(
    net,
    seed,
):
    s_w_net = net.groupby(['source'])['weight'].apply(lambda x: set(sorted(x)))
    s_t_net = net.groupby(['source'])['target'].apply(lambda x: set(sorted(x)))
    t_net = dc.pp.shuffle_net(net=net, target=True, weight=False, same_seed=False, seed=seed)
    s_t_t_net = t_net.groupby(['source'])['target'].apply(lambda x: set(sorted(x)))
    s_w_t_net = t_net.groupby(['source'])['weight'].apply(lambda x: set(sorted(x)))
    inter = s_w_net.index.intersection(s_w_t_net.index)
    assert inter.size > 0
    assert not all(s_t_t_net.loc[s].issubset(s_t_net.loc[s]) for s in inter)
    assert all(s_w_t_net.loc[s].issubset(s_w_net.loc[s]) for s in inter)
    w_net = dc.pp.shuffle_net(net=net, target=False, weight=True, same_seed=False, seed=seed)
    s_t_w_net = w_net.groupby(['source'])['target'].apply(lambda x: set(sorted(x)))
    s_w_w_net = w_net.groupby(['source'])['weight'].apply(lambda x: set(sorted(x)))
    inter = s_w_net.index.intersection(s_w_w_net.index)
    assert inter.size > 0
    assert all(s_t_w_net.loc[s].issubset(s_t_net.loc[s]) for s in inter)
    assert not all(s_w_w_net.loc[s].issubset(s_w_net.loc[s]) for s in inter)
    tw_net = dc.pp.shuffle_net(net=net, target=True, weight=True, same_seed=False, seed=seed)
    stw_net = dc.pp.shuffle_net(net=net, target=True, weight=True, same_seed=True, seed=seed)
    t_net = net.groupby(['target'])['weight'].apply(lambda x: set(sorted(x)))
    t_tw_net = tw_net.groupby(['target'])['weight'].apply(lambda x: set(sorted(x)))
    t_stw_net = stw_net.groupby(['target'])['weight'].apply(lambda x: set(sorted(x)))
    inter = t_net.index.intersection(t_tw_net.index).intersection(t_stw_net.index)
    assert inter.size > 0
    assert not all(t_tw_net.loc[s].issubset(t_net.loc[s]) for s in inter)
    assert all(t_stw_net.loc[s].issubset(t_net.loc[s]) for s in inter)


def test_net_corr(
    adata,
    net,
):
    n_src = net['source'].unique().size
    corr = dc.pp.net_corr(net=net, data=None, tmin=0)
    assert isinstance(corr, pd.DataFrame)
    cols = {'source_a', 'source_b', 'corr', 'pval', 'padj'}
    assert cols.issubset(corr.columns)
    assert (corr['source_a'] != corr['source_b']).all()
    n_pairs = n_src * (n_src - 1) // 2
    assert corr.shape[0] == n_pairs
    corr = dc.pp.net_corr(net=net, data=adata, tmin=0)
    assert isinstance(corr, pd.DataFrame)
    n_pairs = n_src * (n_src - 1) // 2
    assert corr.shape[0] == n_pairs



import pandas as pd
import numpy as np


def _format(
    df: pd.DataFrame,
    cols: list,
) -> pd.DataFrame:
    # Validate
    assert isinstance(df, pd.DataFrame), 'df must be pandas.DataFrame'
    assert isinstance(cols, list), 'cols must be list'
    assert 'metric' in df.columns, 'metric must be in df.columns'
    assert 'score' in df.columns, 'score must be in df.columns'
    # Extract
    tmp = df[df['metric'].isin(cols)].copy()
    assert tmp.shape[0] > 0, 'cols must be in df["metric"]'
    # Add small variations so not same number
    rng = np.random.default_rng(seed=0)
    tmp.loc[:, 'score'] = tmp.loc[:, 'score'] + rng.normal(loc=0, scale=2.2e-16, size=tmp.shape[0])
    tmp.loc[:, 'score'] = tmp.loc[:, 'score'].clip(lower=0)
    # Transform
    grp_cols = ['net', 'groupby', 'group', 'source', 'method']
    grp_cols = [c for c in grp_cols if c in df.columns]
    tmp = (
        tmp
        .pivot(index=grp_cols, columns='metric', values='score')
        .reset_index()
    ).dropna(axis=1)
    if np.all(np.isin(['groupby', 'group'], tmp.columns)):
        tmp = (
            tmp
            .pivot(index=['source', 'method'] + cols, columns='groupby', values='group')
            .reset_index()
        )
    # Remove names
    tmp.index.name = None
    tmp.columns.name = None
    return tmp



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'stat,name',
    [
        ['stat', 'T1'],
        ['stat', 'T2'],
        ['pval', 'T3'],
        ['pval', 'T4'],
    ]
)
def test_leading_edge(
    net,
    stat,
    name,
):
    df = pd.DataFrame(
        data=[[i, i ** 2] for i in range(9)],
        columns=['stat', 'pval'],
        index=[f'G0{i}' for i in range(9)],
    )
    fig, le = dc.pl.leading_edge(df=df, net=net, stat=stat, name=name, return_fig=True)
    assert isinstance(le, np.ndarray)
    assert isinstance(fig, Figure)
    plt.close(fig)



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns
from anndata import AnnData

from decoupler._docs import docs
from decoupler._Plotter import Plotter


@docs.dedent
def obsbar(
    adata: AnnData,
    y: str,
    hue: str | None = None,
    kw_barplot: dict = dict(),
    **kwargs
) -> None | Figure:
    """
    Plot ``adata.obs`` metadata as a grouped barplot.

    Parameters
    ----------
    %(adata)s
    y
        Column name in ``adata.obs`` to plot in y axis.
    hue
        Column name in ``adata.obs`` to color bars.
    kw_barplot
        Keyword arguments passed to ``seaborn.barplot``.
    %(plot)s
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be an AnnData instance'
    assert isinstance(y, str), 'y must be str'
    assert isinstance(hue, str) or hue is None, 'hue must be str or None'
    cols = {y, hue}
    if hue is None:
        cols.remove(None)
    assert cols.issubset(adata.obs.columns), \
    f'y={y} and hue={hue} must be in adata.obs.columns={adata.obs.columns}'
    cols = list(cols)
    # Process
    data = (
        adata.obs
        .groupby(cols, observed=True, as_index=False)
        .size()
    )
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    sns.barplot(
        data=data,
        y=y,
        x='size',
        hue=hue,
        ax=bp.ax,
        **kw_barplot
    )
    if hue is not None and y != hue:
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=hue)
    return bp._return()



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.colors import TwoSlopeNorm

from decoupler._docs import docs
from decoupler._Plotter import Plotter


@docs.dedent
def dotplot(
    df: pd.DataFrame,
    x: str,
    y: str,
    c: str,
    s: str,
    top: int | float = 10,
    scale: int | float = 0.15,
    cmap: str = 'RdBu_r',
    vcenter: int | float | None = None,
    **kwargs
) -> None | Figure:
    """
    Plot results of enrichment analysis as dots.

    Parameters
    ----------
    df
        DataFrame containing enrichment results.
    x
        Name of the column containing values to place on the x-axis.
    y
        Name of the column containing values to place on the y-axis.
    c
        Name of the column containing values to use for coloring.
    s
        Name of the column containing values to use for setting the size of the dots.
    %(top)s
    scale
        Scale of the dots.
    %(cmap)s
    %(vcenter)s
    %(plot)s
    """
    # Validate
    assert isinstance(df, pd.DataFrame), 'df must be a pd.DataFrame'
    assert isinstance(x, str) and x in df.columns, 'x must be str and in df.columns'
    assert isinstance(y, str) and y in df.columns, 'y must be str and in df.columns'
    assert isinstance(c, str) and c in df.columns, 'c must be str and in df.columns'
    assert isinstance(s, str) and s in df.columns, 's must be str and in df.columns'
    assert isinstance(top, (int, float)) and top > 0, 'top must be numerical and > 0'
    assert isinstance(scale, (int, float)), 'scale must be numerical'
    assert isinstance(vcenter, (int, float)) or vcenter is None, 'vcenter must be numeric or None'
    # Filter by top
    df = df.copy()
    df['abs_x_col'] = df[x].abs()
    df = df.sort_values('abs_x_col', ascending=False).head(top)
    # Extract from df
    x_vals = df[x].values
    y_vals = df[y].values
    c_vals = df[c].values
    s_vals = df[s].values
    # Sort by x
    idxs = np.argsort(x_vals)
    x_vals = x_vals[idxs]
    y_vals = y_vals[idxs]
    c_vals = c_vals[idxs]
    s_vals = s_vals[idxs]
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    ns = (s_vals * scale * plt.rcParams["lines.markersize"]) ** 2
    bp.ax.grid(axis='x')
    if vcenter:
        norm = TwoSlopeNorm(vmin=None, vcenter=vcenter, vmax=None)
    else:
        norm = None
    scatter = bp.ax.scatter(
        x=x_vals,
        y=y_vals,
        c=c_vals,
        s=ns,
        cmap=cmap,
        norm=norm,
    )
    bp.ax.set_axisbelow(True)
    bp.ax.set_xlabel(x)
    # Add legend
    handles, labels = scatter.legend_elements(
        prop="sizes",
        num=3,
        fmt="{x:.2f}",
        func=lambda s: np.sqrt(s) / plt.rcParams["lines.markersize"] / scale
    )
    bp.ax.legend(
        handles,
        labels,
        title=s,
        frameon=False,
        loc='lower left',
        bbox_to_anchor=(1.05, 0.5),
        alignment='left',
        labelspacing=1.
    )
    # Add colorbar
    clb = bp.fig.colorbar(
        scatter,
        ax=bp.ax,
        shrink=0.25,
        aspect=5,
        orientation='vertical',
        anchor=(0., 0.),
    )
    clb.ax.set_title(c, loc="left",)
    bp.ax.margins(x=0.25, y=0.1)
    return bp._return()



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'y,hue,kw',
    [
        ['group', None, dict()],
        ['group', 'group', dict(width=0.5)],
        ['group', 'sample', dict(palette='tab10')],
        ['sample', 'group', dict(palette='tab20')],
    ]
)
def test_obsbar(
    adata,
    y,
    hue,
    kw,
):
    fig = dc.pl.obsbar(adata=adata, y=y, hue=hue, kw_barplot=kw, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



import pandas as pd
import numpy as np
import pytest

import decoupler as dc


def test_hallmark():
    hm = dc.op.hallmark()
    assert isinstance(hm, pd.DataFrame)
    cols = {'source', 'target'}
    assert cols.issubset(hm.columns)
    assert not hm.duplicated(['source', 'target']).any()



from typing import Tuple

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from decoupler._docs import docs
from decoupler._Plotter import Plotter


def _set_limits(
    vmin: int | float,
    vcenter: int | float,
    vmax: int | float,
    values: np.ndarray
) -> Tuple[float, float, float]:
    assert np.isfinite(values).all(), 'values in data mut be finite'
    assert isinstance(vmin, (int, float)) or vmin is None, 'vmin must be numerical or None'
    assert isinstance(vcenter, (int, float)) or vcenter is None, 'vcenter must be numerical or None'
    assert isinstance(vmax, (int, float)) or vmax is None, 'vmax must be numerical or None'
    if vmin is None:
        vmin = values.min()
    if vmax is None:
        vmax = values.max()
    if vcenter is None:
        vcenter = values.mean()
    if vmin >= vcenter:
        vmin = -vmax
    if vcenter >= vmax:
        vmax = -vmin
    return vmin, vcenter, vmax


@docs.dedent
def barplot(
    data: pd.DataFrame,
    name: str,
    top: int = 25,
    vertical: bool = False,
    cmap: str = 'RdBu_r',
    vmin: float | None = None,
    vcenter: float | None = 0,
    vmax: float | None = None,
    **kwargs,
) -> None | Figure:
    """
    Plot barplots showing top scores.

    Parameters
    ----------
    data
        DataFrame in wide format containing enrichment scores (contrasts, sources).
    name
        Name of the contrast (row) to plot.
    %(top)s
    vertical
        Whether to plot the bars verticaly or horizontaly.
    %(cmap)s
    %(vmin)s
    %(vcenter)s
    %(vmax)s
    %(plot)s
    """
    # Validate
    assert isinstance(data, pd.DataFrame), 'data must be pandas.DataFrame'
    assert isinstance(name, str) and name in data.index, \
    'name must be str and in data.index'
    assert isinstance(top, int) and top > 0, 'top must be int and > 0'
    assert isinstance(vertical, bool), 'vertical must be bool'
    # Process df
    df = data.loc[[name]]
    df.index.name = None
    df.columns.name = None
    df = df.melt(var_name='source', value_name='score')
    df['abs_score'] = df['score'].abs()
    df = df.sort_values('abs_score', ascending=False)
    df = df.head(top).sort_values('score', ascending=False)
    if not vertical:
        x, y = 'score', 'source'
    else:
        x, y = 'source', 'score'
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    sns.barplot(data=df, x=x, y=y, ax=bp.ax)
    if not vertical:
        sizes = np.array([bar.get_width() for bar in bp.ax.containers[0]])
        bp.ax.set_xlabel('Score')
        bp.ax.set_ylabel('')
    else:
        sizes = np.array([bar.get_height() for bar in bp.ax.containers[0]])
        bp.ax.tick_params(axis='x', rotation=90)
        bp.ax.set_ylabel('Score')
        bp.ax.set_xlabel('')
        bp.ax.invert_xaxis()
    # Compute color limits
    vmin, vcenter, vmax = _set_limits(vmin, vcenter, vmax, df['score'])
    # Rescale cmap
    divnorm = matplotlib.colors.TwoSlopeNorm(vmin=vmin, vcenter=vcenter, vmax=vmax)
    cmap_f = plt.get_cmap(cmap)
    div_colors = cmap_f(divnorm(sizes))
    for bar, color in zip(bp.ax.containers[0], div_colors):
        bar.set_facecolor(color)
    # Add legend
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=divnorm)
    sm.set_array([])
    bp.fig.colorbar(sm, ax=bp.ax, shrink=0.5)
    return bp._return()



from decoupler.ds._bulk import hsctgfb, knocktf
from decoupler.ds._scell import pbmc3k, covid5k, erygast1k
from decoupler.ds._spatial import msvisium
from decoupler.ds._toy import toy, toy_bench
from decoupler.ds._utils import ensmbl_to_symbol



import pandas as pd
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'url,kwargs',
    [
        [
            ('https://www.ncbi.nlm.nih.gov/geo/download/?' +
             'acc=GSM8563697&format=file&file=GSM8563697%' +
             '5FCO37%5Ffeatures%2Etsv%2Egz'),
            dict(sep='\t', compression='gzip', header=None)
        ],
        [
            ('https://www.ncbi.nlm.nih.gov/geo/download/?' +
             'acc=GSM8563697&format=file&file=GSM8563697%' +
             '5FCO37%5Ftissue%5Fpositions%5Flist%2Ecsv%2Egz'),
            dict(sep=',', compression='gzip')
        ], 
    ]
)
def test_download(
    url,
    kwargs,
):
    df = dc._download._download(url, **kwargs)
    assert isinstance(df, pd.DataFrame)
    assert df.columns.size > 1



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns
from anndata import AnnData

from decoupler._docs import docs
from decoupler._Plotter import Plotter


@docs.dedent
def filter_samples(
    adata: AnnData,
    groupby: str | list,
    log: bool = True,
    min_cells: int | float = 10,
    min_counts: int | float = 1000,
    **kwargs
) -> None | Figure:
    """
    Plot to assess the quality of the obtained pseudobulk samples from ``decoupler.pp.pseudobulk``.

    Parameters
    ----------
    %(adata)s
    groupby
        Name or nomes of the ``adata.obs`` column/s to group by.
    log
        If set, log10 transform the ``psbulk_n_cells`` and ``psbulk_counts`` columns during visualization.
    %(min_cells)s
    %(min_counts)s
    %(plot)s
    """
    # Validate
    assert isinstance(adata, AnnData), 'adata must be AnnData'
    assert isinstance(adata.obs, pd.DataFrame) and adata.obs is not None, \
    f'adata.obs must be a pd.DataFrame not {type(adata.obs)}'
    assert all(col in adata.obs.columns for col in ['psbulk_cells', 'psbulk_counts']), \
    'psbulk_* columns not present in adata.obs, this function should be used after running decoupler.pp.pseudobulk'
    assert isinstance(groupby, (str, list)), 'groupby must be str or list'
    if isinstance(groupby, str):
        groupby = [groupby]
    assert all(col in adata.obs for col in groupby), 'columns in groupby must be in adata.obs'
    # Extract obs
    df = adata.obs.copy()
    # Transform to log10
    label_x, label_y = 'cells', 'counts'
    if log:
        df['psbulk_cells'] = np.log10(df['psbulk_cells'] + 1)
        df['psbulk_counts'] = np.log10(df['psbulk_counts'] + 1)
        label_x, label_y = r'$\log_{10}$ ' + label_x, r'$\log_{10}$ ' + label_y
        min_cells, min_counts = np.log10(min_cells), np.log10(min_counts)
    # Plot
    if len(groupby) > 1:
        # Instance
        assert kwargs.get('ax') is None, 'when groupby is list, ax must be None'
        kwargs['ax'] = None
        bp = Plotter(**kwargs)
        bp.fig.delaxes(bp.ax)
        plt.close(bp.fig)
        bp.fig, axes = plt.subplots(len(groupby), 1, figsize=bp.figsize, dpi=bp.dpi, tight_layout=True)
        axes = axes.ravel()
        for ax, grp in zip(axes, groupby):
            ax.grid(zorder=0)
            ax.set_axisbelow(True)
            sns.scatterplot(x='psbulk_cells', y='psbulk_counts', hue=grp, ax=ax, data=df, zorder=1)
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=grp)
            ax.set_xlabel(label_x)
            ax.set_ylabel(label_y)
            ax.axvline(x=min_cells, linestyle='--', color="black")
            ax.axhline(y=min_counts, linestyle='--', color="black")
    else:
        # Instance
        groupby = groupby[0]
        bp = Plotter(**kwargs)
        bp.ax.grid(zorder=0)
        bp.ax.set_axisbelow(True)
        sns.scatterplot(x='psbulk_cells', y='psbulk_counts', hue=groupby, ax=bp.ax, data=df, zorder=1)
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=groupby)
        bp.ax.set_xlabel(label_x)
        bp.ax.set_ylabel(label_y)
        bp.ax.axvline(x=min_cells, linestyle='--', color="black")
        bp.ax.axhline(y=min_counts, linestyle='--', color="black")
    return bp._return()



import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'kwargs',
    [
        dict(),
        dict(n_estimators=10),
        dict(max_depth=1),
        dict(gamma=0.01),
    ]
)
def test_func_udt(
    mat,
    adjmat,
    kwargs,
):
    X, obs, var = mat
    es = dc.mt._udt._func_udt(mat=X, adj=adjmat, **kwargs)[0]
    assert np.isfinite(es).all()
    assert ((0 <= es) & (es <= 1)).all()



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'groupby,log',
    [
        ['group', True],
        [['group'], True],
        [['sample', 'group'], True],
    ]
)
def test_filter_samples(
    pdata,
    groupby,
    log,
):
    fig = dc.pl.filter_samples(adata=pdata, groupby=groupby, log=log, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)



import numpy as np
import pytest

import decoupler as dc


@pytest.mark.parametrize(
    'flavor', ['KSEA', 'RoKAI']
)
def test_func_zscore(
    mat,
    adjmat,
    flavor,
):
    X, obs, var = mat
    es, pv = dc.mt._zscore._func_zscore(mat=X, adj=adjmat, flavor=flavor)
    assert np.isfinite(es).all()
    assert ((0 <= pv) & (pv <= 1)).all()



import numpy as np
import pandas as pd

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._download import URL_INT, _download
from decoupler.op._translate import translate
from decoupler.op._dtype import _infer_dtypes


@docs.dedent
def collectri(
    organism: str = 'human',
    remove_complexes: bool = False,
    license: str = 'academic',
    verbose: bool = False,
) -> pd.DataFrame:
    """
    CollecTRI gene regulatory network :cite:p:`collectri`.

    Wrapper to access CollecTRI gene regulatory network. CollecTRI is a
    comprehensive resource containing a curated collection of transcription
    factors (TFs) and their target genes. It is an expansion of DoRothEA.
    Each interaction is weighted by its mode of regulation (either positive or negative).

    Parameters
    ----------
    %(organism)s
    remove_complexes
        Whether to remove complexes.
    %(license)s
    %(verbose)s

    Returns
    -------
    Dataframe in long format containing target genes for each TF with their associated weights,
    and if available, the PMIDs supporting each interaction.
    """
    url = 'https://zenodo.org/records/8192729/files/CollecTRI_regulons.csv?download=1'
    ct = _download(url, verbose=verbose)
    # Update resources
    resources = []
    for str_res in ct['resources']:
        lst_res = str_res.replace('CollecTRI', '').split(';')
        str_res = ';'.join(sorted([res.replace('_', '') for res in lst_res if res != '']))
        resources.append(str_res)
    ct['resources'] = resources
    # Format references
    ct['references'] = ct['references'].str.replace('CollecTRI:', '')
    ct = ct.dropna()
    if remove_complexes:
        ct = ct[~ct['source'].isin(['AP1', 'NFKB'])]
    ct = _infer_dtypes(ct)
    if organism != 'human':
        ct = translate(ct, columns=['source', 'target'], target_organism=organism, verbose=verbose)
    ct = ct.drop_duplicates(['source', 'target']).reset_index(drop=True)
    return ct



import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.bm.pl._format import _format


@docs.dedent
def auc(
    df: pd.DataFrame,
    hue: str | None = None,
    palette: str = 'tab20',
    thr_auroc: float = 0.5,
    thr_auprc: float = 0.5,
    **kwargs
) -> None | Figure:
    """
    Plot auroc and auprc.

    x-axis represent the auroc calculated by ranking all obtained enrichment scores, calculating different class thresholds
    and finally obtaining the area under the curve.
    The higher value the better performance is.

    y-axis represent the auprc calculated by ranking all obtained enrichment scores, calculating different class thresholds
    and finally obtaining the area under the curve.
    The higher value the better performance is.

    Parameters
    ----------
    %(df)s
    %(hue)s
    %(palette)s
    thr_auroc
        Dashed line to indicate baseline of auroc.
    thr_auprc
        Dashed line to indicate baseline of auprc.
    %(plot)s
    """
    # Validate
    assert isinstance(hue, str) or hue is None, 'hue must be str or None'
    assert isinstance(thr_auroc, float) and 0. <= thr_auroc <= 1., \
    'thr_auroc must be float and between 0 and 1'
    assert isinstance(thr_auprc, float) and 0. <= thr_auprc <= 1., \
    'thr_auprc must be float and between 0 and 1'
    # Format
    tmp = _format(df=df, cols=['auroc', 'auprc'])
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    if hue is not None:
        sns.scatterplot(
            data=tmp,
            x='auroc',
            y='auprc',
            hue=hue,
            ax=bp.ax,
            palette=palette,
        )
    else:
        sns.scatterplot(
            data=tmp,
            x='auroc',
            y='auprc',
            ax=bp.ax,
        )
    bp.ax.axvline(x=thr_auroc, ls='--', c='black', zorder=0)
    bp.ax.axhline(y=thr_auprc, ls='--', c='black', zorder=0)
    if hue is not None:
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=hue)
    return bp._return()



from typing import Tuple

import numpy as np
import scipy.stats as sts

from decoupler._docs import docs
from decoupler._log import _log
from decoupler._Method import MethodMeta, Method


@docs.dedent
def _func_zscore(
    mat: np.ndarray,
    adj: np.ndarray,
    flavor: str = 'RoKAI',
    verbose: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Z-score (ZSCORE) :cite:`zscore`.

    %(yestest)s

    %(params)s

    flavor
        Which flavor to use when calculating the z-score, either KSEA or RoKAI.

    %(returns)s
    """
    assert isinstance(flavor, str) and flavor in ['KSEA', 'RoKAI'], \
    'flavor must be str and KSEA or RoKAI'
    nobs, nvar = mat.shape
    nvar, nsrc = adj.shape
    m = f'zscore - calculating {nsrc} scores with flavor={flavor}'
    _log(m, level='info', verbose=verbose)
    stds = np.std(mat, axis=1, ddof=1)
    if flavor == 'RoKAI':
        mean_all = np.mean(mat, axis=1)
    elif flavor == 'KSEA':
        mean_all = np.zeros(stds.shape)
    n = np.sqrt(np.count_nonzero(adj, axis=0))
    mean = mat.dot(adj) / np.sum(np.abs(adj), axis=0)
    es = ((mean - mean_all.reshape(-1, 1)) * n) / stds.reshape(-1, 1)
    pv = sts.norm.cdf(-np.abs(es))
    return es, pv


_zscore = MethodMeta(
    name='zscore',
    desc='Z-score (ZSCORE)',
    func=_func_zscore,
    stype='numerical',
    adj=True,
    weight=True,
    test=True,
    limits=(-np.inf, +np.inf),
    reference='https://doi.org/10.1038/s41467-021-21211-6',
)
zscore = Method(_method=_zscore)



import pandas as pd
import numpy as np
import seaborn as sns
from matplotlib.figure import Figure

from decoupler._docs import docs
from decoupler._Plotter import Plotter
from decoupler.bm.pl._format import _format


@docs.dedent
def qrank(
    df: pd.DataFrame,
    hue: str | None = None,
    palette: str = 'tab20',
    thr_rank: float = 0.5,
    thr_pval: float = 0.05,
    **kwargs
) -> None | Figure:
    """
    Plot 1-qrank and p-value.

    x-axis represent the one minus the quantile normalized ranks for the sources that belong to the ground truth.
    The closer to 1 the better performance is.

    y-axis represents the p-value (-log10) obtained after performing a Ranksums test between the quantile normalized
    ranks of the sources that belong to the ground truth against the sources that do not.
    The higher value the better performance is.

    Parameters
    ----------
    %(df)s
    %(hue)s
    %(palette)s
    thr_rank
        Dashed line to indicate baseline of ranks.
    thr_pval
        Dashed line to indicate baseline of p-values.
    %(plot)s
    """
    # Validate
    assert isinstance(hue, str) or hue is None, 'hue must be str or None'
    assert isinstance(thr_rank, float) and 0. <= thr_rank <= 1., \
    'thr_rank must be float and between 0 and 1'
    assert isinstance(thr_pval, float) and 0. <= thr_pval <= 1., \
    'thr_pval must be float and between 0 and 1'
    # Format
    tmp = _format(df=df, cols=['1-qrank', '-log10(pval)'])
    # Instance
    bp = Plotter(**kwargs)
    # Plot
    if hue is not None:
        sns.scatterplot(
            data=tmp,
            x='1-qrank',
            y='-log10(pval)',
            hue=hue,
            ax=bp.ax,
            palette=palette,
        )
    else:
        sns.scatterplot(
            data=tmp,
            x='1-qrank',
            y='-log10(pval)',
            ax=bp.ax,
        )
    bp.ax.set_xlim(0, 1)
    bp.ax.axvline(x=thr_rank, ls='--', c='black', zorder=0)
    bp.ax.axhline(y=-np.log10(thr_pval), ls='--', c='black', zorder=0)
    bp.ax.set_ylabel(r'$\log_{10}$(pval)')
    if hue is not None:
        bp.ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), frameon=False, title=hue)
    return bp._return()



import pandas as pd
import pytest

import decoupler as dc

    
def test_infer_dtypes():
    df = pd.DataFrame(
        data = [
            ['1', 'A', 'true', 'False', 0.3],
            ['2', 'B', 'false', 'True', 0.1],
            ['3', 'C', 'false', 'True', 3.1],
        ],
        columns=['a', 'b', 'c', 'd', 'e'],
        index=[0, 1, 2],
    )
    df['b'] = df['b'].astype('string')
    idf = dc.op._dtype._infer_dtypes(df.copy())
    assert pd.api.types.is_numeric_dtype(idf['a'])
    assert idf['b'].dtype == 'object'
    assert pd.api.types.is_bool_dtype(idf['c'])
    assert pd.api.types.is_bool_dtype(idf['d'])
    assert pd.api.types.is_numeric_dtype(idf['e'])

    


import matplotlib.pyplot as plt
from matplotlib.figure import Figure

import decoupler as dc


def test_filter_by_expr(
    pdata,
):
    fig = dc.pl.filter_by_expr(adata=pdata, return_fig=True)
    assert isinstance(fig, Figure)
    plt.close(fig)
